@@ -37,9 +37,9 @@ struct njs_xml_nset_s {
 };
 
 static njs_int_t njs_xml_ext_parse(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused);
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
 static njs_int_t njs_xml_ext_canonicalization(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t magic);
+    njs_uint_t nargs, njs_index_t magic, njs_value_t *retval);
 static njs_int_t njs_xml_doc_ext_prop_keys(njs_vm_t *vm, njs_value_t *value,
     njs_value_t *keys);
 static njs_int_t njs_xml_doc_ext_root(njs_vm_t *vm, njs_object_prop_t *prop,
@@ -55,7 +55,7 @@ static njs_int_t njs_xml_attr_ext_prop_handler(njs_vm_t *vm,
     njs_object_prop_t *prop, njs_value_t *value, njs_value_t *unused,
     njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_add_child(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused);
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_attrs(njs_vm_t *vm, njs_object_prop_t *prop,
     njs_value_t *value, njs_value_t *setval, njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_name(njs_vm_t *vm, njs_object_prop_t *prop,
@@ -65,17 +65,23 @@ static njs_int_t njs_xml_node_ext_ns(njs_vm_t *vm, njs_object_prop_t *prop,
 static njs_int_t njs_xml_node_ext_parent(njs_vm_t *vm, njs_object_prop_t *prop,
      njs_value_t *value, njs_value_t *setval, njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_remove_all_attributes(njs_vm_t *vm,
-    njs_value_t *args, njs_uint_t nargs, njs_index_t unused);
+    njs_value_t *args, njs_uint_t nargs, njs_index_t unused,
+    njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_remove_attribute(njs_vm_t *vm,
-    njs_value_t *args, njs_uint_t nargs, njs_index_t unused);
+    njs_value_t *args, njs_uint_t nargs, njs_index_t unused,
+    njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_remove_children(njs_vm_t *vm,
-    njs_value_t *args, njs_uint_t nargs, njs_index_t unused);
+    njs_value_t *args, njs_uint_t nargs, njs_index_t unused,
+    njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_remove_text(njs_vm_t *vm,
-    njs_value_t *args, njs_uint_t nargs, njs_index_t unused);
+    njs_value_t *args, njs_uint_t nargs, njs_index_t unused,
+    njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_set_attribute(njs_vm_t *vm,
-    njs_value_t *args, njs_uint_t nargs, njs_index_t unused);
+    njs_value_t *args, njs_uint_t nargs, njs_index_t unused,
+    njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_set_text(njs_vm_t *vm,
-    njs_value_t *args, njs_uint_t nargs, njs_index_t unused);
+    njs_value_t *args, njs_uint_t nargs, njs_index_t unused,
+    njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_tags(njs_vm_t *vm, njs_object_prop_t *prop,
      njs_value_t *value, njs_value_t *setval, njs_value_t *retval);
 static njs_int_t njs_xml_node_ext_text(njs_vm_t *vm, njs_object_prop_t *prop,
@@ -396,6 +402,7 @@ static njs_external_t  njs_ext_xml_attr[] = {
 
 njs_module_t  njs_xml_module = {
     .name = njs_str("xml"),
+    .preinit = NULL,
     .init = njs_xml_init,
 };
 
@@ -407,7 +414,7 @@ static njs_int_t    njs_xml_attr_proto_id;
 
 static njs_int_t
 njs_xml_ext_parse(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t         ret;
     njs_str_t         data;
@@ -427,7 +434,7 @@ njs_xml_ext_parse(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
     tree->ctx = xmlNewParserCtxt();
     if (njs_slow_path(tree->ctx == NULL)) {
-        njs_vm_error(vm, "xmlNewParserCtxt() failed");
+        njs_vm_internal_error(vm, "xmlNewParserCtxt() failed");
         return NJS_ERROR;
     }
 
@@ -448,7 +455,7 @@ njs_xml_ext_parse(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     cln->handler = njs_xml_doc_cleanup;
     cln->data = tree;
 
-    return njs_vm_external_create(vm, njs_vm_retval(vm), njs_xml_doc_proto_id,
+    return njs_vm_external_create(vm, retval, njs_xml_doc_proto_id,
                                   tree, 0);
 }
 
@@ -520,6 +527,11 @@ njs_xml_doc_ext_root(njs_vm_t *vm, njs_object_prop_t *prop, njs_value_t *value,
             njs_value_undefined_set(retval);
             return NJS_DECLINED;
         }
+
+    } else {
+        /* To suppress warning. */
+        name.length = 0;
+        name.start = NULL;
     }
 
     for (node = xmlDocGetRootElement(tree->doc);
@@ -705,49 +717,49 @@ njs_xml_node_ext_prop_handler(njs_vm_t *vm, njs_object_prop_t *prop,
 
 static njs_int_t
 njs_xml_node_ext_add_child(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     xmlNode    *current, *node, *copy, *rnode;
     njs_int_t  ret;
 
     current = njs_vm_external(vm, njs_xml_node_proto_id, njs_argument(args, 0));
     if (njs_slow_path(current == NULL)) {
-        njs_vm_error(vm, "\"this\" is not a XMLNode object");
+        njs_vm_type_error(vm, "\"this\" is not a XMLNode object");
         return NJS_ERROR;
     }
 
     node = njs_xml_external_node(vm, njs_arg(args, nargs, 1));
     if (njs_slow_path(node == NULL)) {
-        njs_vm_error(vm, "node is not a XMLNode object");
+        njs_vm_type_error(vm, "node is not a XMLNode object");
         return NJS_ERROR;
     }
 
     copy = xmlDocCopyNode(current, current->doc, 1);
     if (njs_slow_path(copy == NULL)) {
-        njs_vm_error(vm, "xmlDocCopyNode() failed");
+        njs_vm_internal_error(vm, "xmlDocCopyNode() failed");
         return NJS_ERROR;
     }
 
     node = xmlDocCopyNode(node, current->doc, 1);
     if (njs_slow_path(node == NULL)) {
-        njs_vm_error(vm, "xmlDocCopyNode() failed");
+        njs_vm_internal_error(vm, "xmlDocCopyNode() failed");
         goto error;
     }
 
     rnode = xmlAddChild(copy, node);
     if (njs_slow_path(rnode == NULL)) {
         xmlFreeNode(node);
-        njs_vm_error(vm, "xmlAddChild() failed");
+        njs_vm_internal_error(vm, "xmlAddChild() failed");
         goto error;
     }
 
     ret = xmlReconciliateNs(current->doc, copy);
     if (njs_slow_path(ret == -1)) {
-        njs_vm_error(vm, "xmlReconciliateNs() failed");
+        njs_vm_internal_error(vm, "xmlReconciliateNs() failed");
         goto error;
     }
 
-    njs_value_undefined_set(njs_vm_retval(vm));
+    njs_value_undefined_set(retval);
 
     return njs_xml_replace_node(vm, current, copy);
 
@@ -832,21 +844,21 @@ njs_xml_node_ext_parent(njs_vm_t *vm, njs_object_prop_t *prop,
 
 static njs_int_t
 njs_xml_node_ext_remove_attribute(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
-    return njs_xml_node_ext_set_attribute(vm, args, nargs, 1);
+    return njs_xml_node_ext_set_attribute(vm, args, nargs, 1, retval);
 }
 
 
 static njs_int_t
 njs_xml_node_ext_remove_all_attributes(njs_vm_t *vm,
-    njs_value_t *args, njs_uint_t nargs, njs_index_t unused)
+    njs_value_t *args, njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     xmlNode  *current;
 
     current = njs_vm_external(vm, njs_xml_node_proto_id, njs_argument(args, 0));
     if (njs_slow_path(current == NULL)) {
-        njs_vm_error(vm, "\"this\" is not a XMLNode object");
+        njs_vm_type_error(vm, "\"this\" is not a XMLNode object");
         return NJS_ERROR;
     }
 
@@ -855,7 +867,7 @@ njs_xml_node_ext_remove_all_attributes(njs_vm_t *vm,
         current->properties = NULL;
     }
 
-    njs_value_undefined_set(njs_vm_retval(vm));
+    njs_value_undefined_set(retval);
 
     return NJS_OK;
 }
@@ -863,7 +875,7 @@ njs_xml_node_ext_remove_all_attributes(njs_vm_t *vm,
 
 static njs_int_t
 njs_xml_node_ext_remove_children(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     xmlNode      *current, *copy;
     njs_str_t    name;
@@ -871,17 +883,17 @@ njs_xml_node_ext_remove_children(njs_vm_t *vm, njs_value_t *args,
 
     current = njs_vm_external(vm, njs_xml_node_proto_id, njs_argument(args, 0));
     if (njs_slow_path(current == NULL)) {
-        njs_vm_error(vm, "\"this\" is not a XMLNode object");
+        njs_vm_type_error(vm, "\"this\" is not a XMLNode object");
         return NJS_ERROR;
     }
 
     selector = njs_arg(args, nargs, 1);
 
-    njs_value_undefined_set(njs_vm_retval(vm));
+    njs_value_undefined_set(retval);
 
     if (!njs_value_is_null_or_undefined(selector)) {
         if (njs_slow_path(!njs_value_is_string(selector))) {
-            njs_vm_error(vm, "selector is not a string");
+            njs_vm_type_error(vm, "selector is not a string");
             return NJS_ERROR;
         }
 
@@ -894,7 +906,7 @@ njs_xml_node_ext_remove_children(njs_vm_t *vm, njs_value_t *args,
 
     copy = xmlDocCopyNode(current, current->doc, 1);
     if (njs_slow_path(copy == NULL)) {
-        njs_vm_error(vm, "xmlDocCopyNode() failed");
+        njs_vm_internal_error(vm, "xmlDocCopyNode() failed");
         return NJS_ERROR;
     }
 
@@ -909,7 +921,7 @@ njs_xml_node_ext_remove_children(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_xml_node_ext_remove_text(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     return njs_xml_node_ext_text(vm, NULL, njs_argument(args, 0), NULL, NULL);
 }
@@ -917,7 +929,7 @@ njs_xml_node_ext_remove_text(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_xml_node_ext_set_attribute(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t remove)
+    njs_uint_t nargs, njs_index_t remove, njs_value_t *retval)
 {
     xmlNode      *current;
     njs_str_t    str;
@@ -925,30 +937,30 @@ njs_xml_node_ext_set_attribute(njs_vm_t *vm, njs_value_t *args,
 
     current = njs_vm_external(vm, njs_xml_node_proto_id, njs_argument(args, 0));
     if (njs_slow_path(current == NULL)) {
-        njs_vm_error(vm, "\"this\" is not a XMLNode object");
+        njs_vm_type_error(vm, "\"this\" is not a XMLNode object");
         return NJS_ERROR;
     }
 
     name = njs_arg(args, nargs, 1);
 
     if (njs_slow_path(!njs_value_is_string(name))) {
-        njs_vm_error(vm, "name is not a string");
+        njs_vm_type_error(vm, "name is not a string");
         return NJS_ERROR;
     }
 
     njs_value_string_get(name, &str);
 
     return njs_xml_node_attr_handler(vm, current, &str, njs_arg(args, nargs, 2),
-                                     !remove ? njs_vm_retval(vm) : NULL);
+                                     !remove ? retval : NULL);
 }
 
 
 static njs_int_t
 njs_xml_node_ext_set_text(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     return njs_xml_node_ext_text(vm, NULL, njs_argument(args, 0),
-                                 njs_arg(args, nargs, 1), njs_vm_retval(vm));
+                                 njs_arg(args, nargs, 1), retval);
 }
 
 
@@ -1005,7 +1017,7 @@ njs_xml_node_ext_text(njs_vm_t *vm, njs_object_prop_t *unused,
         && (setval != NULL && !njs_value_is_null_or_undefined(setval)))
     {
         if (njs_slow_path(!njs_value_is_string(setval))) {
-            njs_vm_error(vm, "setval is not a string");
+            njs_vm_type_error(vm, "setval is not a string");
             return NJS_ERROR;
         }
 
@@ -1019,7 +1031,7 @@ njs_xml_node_ext_text(njs_vm_t *vm, njs_object_prop_t *unused,
 
     copy = xmlDocCopyNode(current, current->doc, 1);
     if (njs_slow_path(copy == NULL)) {
-        njs_vm_error(vm, "xmlDocCopyNode() failed");
+        njs_vm_internal_error(vm, "xmlDocCopyNode() failed");
         return NJS_ERROR;
     }
 
@@ -1084,7 +1096,7 @@ njs_xml_node_attr_handler(njs_vm_t *vm, xmlNode *current, njs_str_t *name,
 
     ret = xmlValidateQName(&name_buf[0], 0);
     if (njs_slow_path(ret != 0)) {
-        njs_vm_error(vm, "attribute name \"%V\" is not valid", name);
+        njs_vm_type_error(vm, "attribute name \"%V\" is not valid", name);
         return NJS_ERROR;
     }
 
@@ -1110,7 +1122,7 @@ njs_xml_node_attr_handler(njs_vm_t *vm, xmlNode *current, njs_str_t *name,
 
     attr = xmlSetProp(current, &name_buf[0], value);
     if (njs_slow_path(attr == NULL)) {
-        njs_vm_error(vm, "xmlSetProp() failed");
+        njs_vm_internal_error(vm, "xmlSetProp() failed");
         return NJS_ERROR;
     }
 
@@ -1129,7 +1141,7 @@ njs_xml_node_tag_remove(njs_vm_t *vm, xmlNode *current, njs_str_t *name)
 
     copy = xmlDocCopyNode(current, current->doc, 1);
     if (njs_slow_path(copy == NULL)) {
-        njs_vm_error(vm, "xmlDocCopyNode() failed");
+        njs_vm_internal_error(vm, "xmlDocCopyNode() failed");
         return NJS_ERROR;
     }
 
@@ -1193,8 +1205,9 @@ njs_xml_node_tag_handler(njs_vm_t *vm, xmlNode *current, njs_str_t *name,
     }
 
     if (retval != NULL) {
-        njs_vm_error(vm, "XMLNode.$tag$xxx is not assignable, use addChild() or"
-                     "node.$tags = [node1, node2, ..] syntax");
+        njs_vm_type_error(vm, "XMLNode.$tag$xxx is not assignable, "
+                          "use addChild() or node.$tags = [node1, node2, ..] "
+                          "syntax");
         return NJS_ERROR;
     }
 
@@ -1254,8 +1267,9 @@ njs_xml_node_tags_handler(njs_vm_t *vm, xmlNode *current, njs_str_t *name,
     }
 
     if (name->length > 0) {
-        njs_vm_error(vm, "XMLNode $tags$xxx is not assignable, use addChild() "
-                     "or node.$tags = [node1, node2, ..] syntax");
+        njs_vm_type_error(vm, "XMLNode $tags$xxx is not assignable, use "
+                          "addChild() or node.$tags = [node1, node2, ..] "
+                          "syntax");
         return NJS_ERROR;
     }
 
@@ -1263,7 +1277,7 @@ njs_xml_node_tags_handler(njs_vm_t *vm, xmlNode *current, njs_str_t *name,
 
     copy = xmlDocCopyNode(current, current->doc, 1);
     if (njs_slow_path(copy == NULL)) {
-        njs_vm_error(vm, "xmlDocCopyNode() failed");
+        njs_vm_internal_error(vm, "xmlDocCopyNode() failed");
         return NJS_ERROR;
     }
 
@@ -1278,7 +1292,7 @@ njs_xml_node_tags_handler(njs_vm_t *vm, xmlNode *current, njs_str_t *name,
     }
 
     if (!njs_value_is_array(setval)) {
-        njs_vm_error(vm, "setval is not an array");
+        njs_vm_type_error(vm, "setval is not an array");
         goto error;
     }
 
@@ -1292,26 +1306,26 @@ njs_xml_node_tags_handler(njs_vm_t *vm, xmlNode *current, njs_str_t *name,
     for (i = 0; i < length; i++) {
         node = njs_xml_external_node(vm, njs_value_arg(start++));
         if (njs_slow_path(node == NULL)) {
-            njs_vm_error(vm, "setval[%D] is not a XMLNode object", i);
+            njs_vm_type_error(vm, "setval[%D] is not a XMLNode object", i);
             goto error;
         }
 
         node = xmlDocCopyNode(node, current->doc, 1);
         if (njs_slow_path(node == NULL)) {
-            njs_vm_error(vm, "xmlDocCopyNode() failed");
+            njs_vm_internal_error(vm, "xmlDocCopyNode() failed");
             goto error;
         }
 
         rnode = xmlAddChild(copy, node);
         if (njs_slow_path(rnode == NULL)) {
-            njs_vm_error(vm, "xmlAddChild() failed");
+            njs_vm_internal_error(vm, "xmlAddChild() failed");
             xmlFreeNode(node);
             goto error;
         }
 
         ret = xmlReconciliateNs(current->doc, copy);
         if (njs_slow_path(ret == -1)) {
-            njs_vm_error(vm, "xmlReconciliateNs() failed");
+            njs_vm_internal_error(vm, "xmlReconciliateNs() failed");
             goto error;
         }
     }
@@ -1338,13 +1352,13 @@ njs_xml_external_node(njs_vm_t *vm, njs_value_t *value)
     if (njs_slow_path(current == NULL)) {
         tree = njs_vm_external(vm, njs_xml_doc_proto_id, value);
         if (njs_slow_path(tree == NULL)) {
-            njs_vm_error(vm, "\"this\" is not a XMLNode object");
+            njs_vm_type_error(vm, "\"this\" is not a XMLNode object");
             return NULL;
         }
 
         current = xmlDocGetRootElement(tree->doc);
         if (njs_slow_path(current == NULL)) {
-            njs_vm_error(vm, "\"this\" is not a XMLNode object");
+            njs_vm_type_error(vm, "\"this\" is not a XMLNode object");
             return NULL;
         }
     }
@@ -1360,8 +1374,8 @@ njs_xml_str_to_c_string(njs_vm_t *vm, njs_str_t *str, u_char *dst,
     u_char  *p;
 
     if (njs_slow_path(str->length > size - njs_length("\0"))) {
-        njs_vm_error(vm, "njs_xml_str_to_c_string() very long string, "
-                     "length >= %uz", size - njs_length("\0"));
+        njs_vm_internal_error(vm, "njs_xml_str_to_c_string() very long string, "
+                              "length >= %uz", size - njs_length("\0"));
         return NJS_ERROR;
     }
 
@@ -1701,7 +1715,7 @@ njs_xml_parse_ns_list(njs_vm_t *vm, njs_str_t *src)
 
 static njs_int_t
 njs_xml_ext_canonicalization(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t magic)
+    njs_index_t magic, njs_value_t *retval)
 {
     u_char           **prefix_list;
     ssize_t          size;
@@ -1726,7 +1740,8 @@ njs_xml_ext_canonicalization(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     if (!njs_value_is_null_or_undefined(excluding)) {
         node = njs_vm_external(vm, njs_xml_node_proto_id, excluding);
         if (njs_slow_path(node == NULL)) {
-            njs_vm_error(vm, "\"excluding\" argument is not a XMLNode object");
+            njs_vm_type_error(vm, "\"excluding\" argument is not a XMLNode "
+                              "object");
             return NJS_ERROR;
         }
 
@@ -1758,7 +1773,7 @@ njs_xml_ext_canonicalization(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
     if (!njs_value_is_null_or_undefined(prefixes)) {
         if (!njs_value_is_string(prefixes)) {
-            njs_vm_error(vm, "\"prefixes\" argument is not a string");
+            njs_vm_type_error(vm, "\"prefixes\" argument is not a string");
             return NJS_ERROR;
         }
 
@@ -1777,7 +1792,7 @@ njs_xml_ext_canonicalization(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
     buf = xmlOutputBufferCreateIO(njs_xml_buf_write_cb, NULL, &chain, NULL);
     if (njs_slow_path(buf == NULL)) {
-        njs_vm_error(vm, "xmlOutputBufferCreateIO() failed");
+        njs_vm_internal_error(vm, "xmlOutputBufferCreateIO() failed");
         return NJS_ERROR;
     }
 
@@ -1786,13 +1801,13 @@ njs_xml_ext_canonicalization(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
                          prefix_list, comments, buf);
 
     if (njs_slow_path(ret < 0)) {
-        njs_vm_error(vm, "xmlC14NExecute() failed");
+        njs_vm_internal_error(vm, "xmlC14NExecute() failed");
         ret = NJS_ERROR;
         goto error;
     }
 
     if (magic & 0x2) {
-        ret = njs_vm_value_string_create_chb(vm, njs_vm_retval(vm), &chain);
+        ret = njs_vm_value_string_create_chb(vm, retval, &chain);
 
     } else {
         size = njs_chb_size(&chain);
@@ -1808,7 +1823,7 @@ njs_xml_ext_canonicalization(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
             goto error;
         }
 
-        ret = njs_vm_value_buffer_set(vm, njs_vm_retval(vm), data.start,
+        ret = njs_vm_value_buffer_set(vm, retval, data.start,
                                       data.length);
     }
 
@@ -1990,7 +2005,7 @@ njs_xml_error(njs_vm_t *vm, njs_xml_doc_t *current, const char *fmt, ...)
                         err->int2);
     }
 
-    njs_vm_value_error_set(vm, njs_vm_retval(vm), "%*s", p - errstr, errstr);
+    njs_vm_error(vm, "%*s", p - errstr, errstr);
 }
 
 
