@@ -325,7 +325,7 @@ njs_module_add(njs_vm_t *vm, njs_str_t *name, njs_value_t *value)
 
 njs_int_t
 njs_module_require(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t    ret;
     njs_str_t    name;
@@ -352,7 +352,7 @@ njs_module_require(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    njs_value_assign(&vm->retval, &module->value);
+    njs_value_assign(retval, &module->value);
 
     return NJS_OK;
 }
