@@ -30,7 +30,7 @@ static njs_int_t njs_object_own_enumerate_object(njs_vm_t *vm,
     const njs_object_t *object, const njs_object_t *parent, njs_array_t *items,
     njs_object_enum_t kind, njs_object_enum_type_t type, njs_bool_t all);
 static njs_int_t njs_object_define_properties(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused);
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
 static njs_int_t njs_object_set_prototype(njs_vm_t *vm, njs_object_t *object,
     const njs_value_t *value);
 
@@ -45,7 +45,7 @@ njs_object_alloc(njs_vm_t *vm)
     if (njs_fast_path(object != NULL)) {
         njs_lvlhsh_init(&object->hash);
         njs_lvlhsh_init(&object->shared_hash);
-        object->__proto__ = &vm->prototypes[NJS_OBJ_TYPE_OBJECT].object;
+        object->__proto__ = njs_vm_proto(vm, NJS_OBJ_TYPE_OBJECT);
         object->slots = NULL;
         object->type = NJS_OBJECT;
         object->shared = 0;
@@ -80,7 +80,7 @@ njs_object_value_copy(njs_vm_t *vm, njs_value_t *value)
 
     if (njs_fast_path(object != NULL)) {
         memcpy(object, njs_object(value), size);
-        object->__proto__ = &vm->prototypes[NJS_OBJ_TYPE_OBJECT].object;
+        object->__proto__ = njs_vm_proto(vm, NJS_OBJ_TYPE_OBJECT);
         object->shared = 0;
         value->data.u.object = object;
         return object;
@@ -119,7 +119,7 @@ njs_object_value_alloc(njs_vm_t *vm, njs_uint_t prototype_index, size_t extra,
     ov->object.error_data = 0;
     ov->object.fast_array = 0;
 
-    ov->object.__proto__ = &vm->prototypes[prototype_index].object;
+    ov->object.__proto__ = njs_vm_proto(vm, prototype_index);
     ov->object.slots = NULL;
 
     if (value != NULL) {
@@ -216,7 +216,7 @@ njs_object_hash_test(njs_lvlhsh_query_t *lhq, void *data)
 
 static njs_int_t
 njs_object_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_uint_t          type, index;
     njs_value_t         *value;
@@ -232,7 +232,7 @@ njs_object_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
             return NJS_ERROR;
         }
 
-        njs_set_object(&vm->retval, object);
+        njs_set_object(retval, object);
 
         return NJS_OK;
     }
@@ -244,7 +244,7 @@ njs_object_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
             return NJS_ERROR;
         }
 
-        njs_set_object_value(&vm->retval, obj_val);
+        njs_set_object_value(retval, obj_val);
 
         return NJS_OK;
     }
@@ -256,7 +256,7 @@ njs_object_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    njs_value_assign(&vm->retval, value);
+    njs_value_assign(retval, value);
 
     return NJS_OK;
 }
@@ -264,7 +264,7 @@ njs_object_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_create(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_value_t   *value, *descs, arguments[3];
     njs_object_t  *object;
@@ -286,18 +286,19 @@ njs_object_create(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
             object->__proto__ = NULL;
         }
 
-        njs_set_object(&vm->retval, object);
-
         descs = njs_arg(args, nargs, 2);
 
         if (njs_slow_path(!njs_is_undefined(descs))) {
             arguments[0] = args[0];
-            arguments[1] = vm->retval;
+            njs_set_object(&arguments[1], object);
             arguments[2] = *descs;
 
-            return njs_object_define_properties(vm, arguments, 3, unused);
+            return njs_object_define_properties(vm, arguments, 3, unused,
+                                                retval);
         }
 
+        njs_set_object(retval, object);
+
         return NJS_OK;
     }
 
@@ -310,7 +311,7 @@ njs_object_create(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_keys(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_value_t  *value;
     njs_array_t  *keys;
@@ -330,7 +331,7 @@ njs_object_keys(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    njs_set_array(&vm->retval, keys);
+    njs_set_array(retval, keys);
 
     return NJS_OK;
 }
@@ -338,7 +339,7 @@ njs_object_keys(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_values(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_array_t  *array;
     njs_value_t  *value;
@@ -358,7 +359,7 @@ njs_object_values(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    njs_set_array(&vm->retval, array);
+    njs_set_array(retval, array);
 
     return NJS_OK;
 }
@@ -366,7 +367,7 @@ njs_object_values(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_entries(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_array_t  *array;
     njs_value_t  *value;
@@ -386,7 +387,7 @@ njs_object_entries(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    njs_set_array(&vm->retval, array);
+    njs_set_array(retval, array);
 
     return NJS_OK;
 }
@@ -871,229 +872,346 @@ njs_object_enumerate_object(njs_vm_t *vm, const njs_object_t *object,
 }
 
 
+#define njs_process_prop(vm, prop, type, items, items_symbol)                  \
+    if (!(type & NJS_ENUM_SYMBOL && njs_is_symbol(&prop->name))) {             \
+        /*                                                                     \
+         * prop from shared_hash is not symbol:                                \
+         * add to items before props from hash                                 \
+         */                                                                    \
+                                                                               \
+        ret = njs_array_add(vm, items, &prop->name);                           \
+        if (njs_slow_path(ret != NJS_OK)) {                                    \
+            return NJS_ERROR;                                                  \
+        }                                                                      \
+                                                                               \
+    } else {                                                                   \
+        /*                                                                     \
+         * prop from shared_hash is symbol:                                    \
+         * add to items_symbol                                                 \
+         */                                                                    \
+        ret = njs_array_add(vm, items_symbol, &prop->name);                    \
+        if (njs_slow_path(ret != NJS_OK)) {                                    \
+            return NJS_ERROR;                                                  \
+        }                                                                      \
+    }
+
+
+static njs_int_t
+njs_get_own_ordered_keys(njs_vm_t *vm, const njs_object_t *object,
+    const njs_object_t *parent, njs_array_t *items, njs_object_enum_t kind,
+    njs_object_enum_type_t type, njs_bool_t all)
+{
+    double              num;
+    uint32_t            items_length;
+    njs_int_t           ret;
+    njs_array_t         *items_string, *items_symbol;
+    njs_lvlhsh_each_t   lhe;
+    njs_object_prop_t   *prop, *ext_prop;
+    njs_lvlhsh_query_t  lhq;
+    const njs_lvlhsh_t  *hash;
+
+    items_length = items->length;
+
+    items_string = njs_array_alloc(vm, 1, 0, NJS_ARRAY_SPARE);
+    if (njs_slow_path(items_string == NULL)) {
+        return NJS_ERROR;
+    }
+
+    items_symbol = njs_array_alloc(vm, 1, 0, NJS_ARRAY_SPARE);
+    if (njs_slow_path(items_symbol == NULL)) {
+        return NJS_ERROR;
+    }
+
+    lhq.proto = &njs_object_hash_proto;
+
+    njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
+    hash = &object->shared_hash;
+
+    for ( ;; ) {
+        prop = njs_lvlhsh_each(hash, &lhe);
+
+        if (prop == NULL) {
+            break;
+        }
+
+        if (!njs_is_enumerable(&prop->name, type)) {
+            continue;
+        }
+
+        njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
+
+        ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
+        if (ext_prop != NULL) {
+            continue;
+        }
+
+        ret = njs_lvlhsh_find(&object->hash, &lhq);
+        if (ret != NJS_OK) {
+
+            if (!(prop->enumerable || all)) {
+                continue;
+            }
+
+            /* prop is:  !in_hash && in_shared_hash */
+
+            num = njs_string_to_index(&prop->name);
+            if (!njs_number_is_integer_index(num)) {
+                njs_process_prop(vm, prop, type, items_string, items_symbol);
+
+            } else {
+                ret = njs_array_add(vm, items, &prop->name);
+                if (njs_slow_path(ret != NJS_OK)) {
+                    return NJS_ERROR;
+                }
+            }
+
+        } else {
+
+            if (!(((njs_object_prop_t *)(lhq.value))->enumerable || all)) {
+                continue;
+            }
+
+            /* prop is:  in_hash && in_shared_hash */
+
+            num = njs_string_to_index(&prop->name);
+            if (!njs_number_is_integer_index(num)) {
+
+                njs_object_prop_t *hash_prop = lhq.value;
+
+                /* select names of prop which are not deleted and
+                 * not deleted and created again i.e.,
+                 * they are replaced shared hash props
+                 */
+                if (hash_prop->type != NJS_WHITEOUT &&
+                    !(hash_prop->enum_in_object_hash))
+                {
+                    njs_process_prop(vm, prop, type, items_string,
+                                     items_symbol);
+                }
+            }
+        }
+    }
+
+    njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
+    hash = &object->hash;
+
+    for ( ;; ) {
+        prop = njs_lvlhsh_each(hash, &lhe);
+
+        if (prop == NULL) {
+            break;
+        }
+
+        if (!njs_is_enumerable(&prop->name, type) ||
+            !(prop->enumerable || all) ||
+            prop->type == NJS_WHITEOUT)
+        {
+            continue;
+        }
+
+        njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
+
+        ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
+        if (ext_prop != NULL) {
+            continue;
+        }
+
+        num = njs_string_to_index(&prop->name);
+        if (njs_number_is_integer_index(num)) {
+
+            ret = njs_array_add(vm, items, &prop->name);
+            if (njs_slow_path(ret != NJS_OK)) {
+                return NJS_ERROR;
+            }
+
+        } else {
+
+            ret = njs_lvlhsh_find(&object->shared_hash, &lhq);
+            if (ret != NJS_OK) {
+                /* prop is:  in_hash && !in_shared_hash */
+
+                /* select names of not deleted props */
+
+                njs_process_prop(vm, prop, type, items_string,
+                                 items_symbol);
+
+            } else {
+                /* prop is:  in_hash && in_shared_hash */
+
+                /* select names of not deleted and created again */
+
+                if (prop->enum_in_object_hash) {
+                    njs_process_prop(vm, prop, type, items_string,
+                                     items_symbol);
+                }
+            }
+        }
+    }
+
+    if (items->length >= 2) {
+        njs_qsort(&items->start[items_length], items->length-items_length,
+                  sizeof(njs_value_t), njs_array_indices_handler_nums, NULL);
+    }
+
+    if (items_string->length != 0) {
+        ret = njs_array_expand(vm, items, 0, items_string->length);
+        if (njs_slow_path(ret != NJS_OK)) {
+            return NJS_ERROR;
+        }
+
+        memcpy(&items->start[items->length], &items_string->start[0],
+               items_string->length * sizeof(njs_value_t));
+
+        items->length += items_string->length;
+    }
+
+    if (items_symbol->length != 0) {
+        ret = njs_array_expand(vm, items, 0, items_symbol->length);
+        if (njs_slow_path(ret != NJS_OK)) {
+            return NJS_ERROR;
+        }
+
+        memcpy(&items->start[items->length], &items_symbol->start[0],
+               items_symbol->length * sizeof(njs_value_t));
+
+        items->length += items_symbol->length;
+    }
+
+    njs_array_destroy(vm, items_string);
+    njs_array_destroy(vm, items_symbol);
+
+    return NJS_OK;
+}
+
+
+static njs_int_t
+njs_add_obj_prop_kind(njs_vm_t *vm, const njs_object_t *object,
+    const njs_lvlhsh_t *hash, njs_lvlhsh_query_t *lhq,
+    njs_object_enum_t kind, njs_array_t *items)
+{
+    njs_int_t          ret;
+    njs_value_t        value, *v, value1;
+    njs_array_t        *entry;
+    njs_object_prop_t  *prop;
+
+    ret = njs_lvlhsh_find(hash, lhq);
+    if (ret != NJS_OK) {
+        return NJS_DECLINED;
+    }
+
+    prop = (njs_object_prop_t *) (lhq->value);
+
+    if (prop->type != NJS_ACCESSOR) {
+        v = njs_prop_value(prop);
+
+    } else {
+        if (njs_is_data_descriptor(prop)) {
+            v = njs_prop_value(prop);
+            goto add;
+        }
+
+        if (njs_prop_getter(prop) == NULL) {
+            v =  njs_value_arg(&njs_value_undefined);
+            goto add;
+        }
+
+        v = &value1;
+
+        njs_set_object(&value, (njs_object_t *) object);
+        ret = njs_function_apply(vm, njs_prop_getter(prop), &value, 1, v);
+        if (ret != NJS_OK) {
+            return NJS_ERROR;
+        }
+    }
+
+add:
+    if (kind != NJS_ENUM_VALUES) {
+        entry = njs_array_alloc(vm, 0, 2, 0);
+        if (njs_slow_path(entry == NULL)) {
+            return NJS_ERROR;
+        }
+
+        njs_string_copy(&entry->start[0], &prop->name);
+        njs_value_assign(&entry->start[1], v);
+
+        njs_set_array(&value, entry);
+        v = &value;
+    }
+
+    ret = njs_array_add(vm, items, v);
+    if (njs_slow_path(ret != NJS_OK)) {
+        return NJS_ERROR;
+    }
+
+    return NJS_OK;
+}
+
+
 static njs_int_t
 njs_object_own_enumerate_object(njs_vm_t *vm, const njs_object_t *object,
     const njs_object_t *parent, njs_array_t *items, njs_object_enum_t kind,
     njs_object_enum_type_t type, njs_bool_t all)
 {
     njs_int_t           ret;
-    njs_value_t         value, *v;
-    njs_array_t         *entry;
+    uint32_t            i;
+    njs_array_t         *items_sorted;
     njs_lvlhsh_each_t   lhe;
-    njs_object_prop_t   *prop, *ext_prop;
     njs_lvlhsh_query_t  lhq;
-    const njs_lvlhsh_t  *hash;
 
     lhq.proto = &njs_object_hash_proto;
 
     njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
-    hash = &object->hash;
 
     switch (kind) {
     case NJS_ENUM_KEYS:
-        for ( ;; ) {
-            prop = njs_lvlhsh_each(hash, &lhe);
-
-            if (prop == NULL) {
-                break;
-            }
-
-            if (!njs_is_enumerable(&prop->name, type)) {
-                continue;
-            }
-
-            njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
-
-            ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
-
-            if (ext_prop == NULL && prop->type != NJS_WHITEOUT
-                && (prop->enumerable || all))
-            {
-                ret = njs_array_add(vm, items, &prop->name);
-                if (njs_slow_path(ret != NJS_OK)) {
-                    return NJS_ERROR;
-                }
-            }
-        }
-
-        njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
-        hash = &object->shared_hash;
-
-        for ( ;; ) {
-            prop = njs_lvlhsh_each(hash, &lhe);
-
-            if (prop == NULL) {
-                break;
-            }
-
-            if (!njs_is_enumerable(&prop->name, type)) {
-                continue;
-            }
-
-            njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
-
-            ret = njs_lvlhsh_find(&object->hash, &lhq);
-
-            if (ret != NJS_OK) {
-                ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
-
-                if (ext_prop == NULL && (prop->enumerable || all)) {
-                    ret = njs_array_add(vm, items, &prop->name);
-                    if (njs_slow_path(ret != NJS_OK)) {
-                        return NJS_ERROR;
-                    }
-                }
-            }
+        ret = njs_get_own_ordered_keys(vm, object, parent, items, kind, type,
+                                       all);
+        if (ret != NJS_OK) {
+            return NJS_ERROR;
         }
 
         break;
 
     case NJS_ENUM_VALUES:
-        for ( ;; ) {
-            prop = njs_lvlhsh_each(hash, &lhe);
-
-            if (prop == NULL) {
-                break;
-            }
-
-            if (!njs_is_enumerable(&prop->name, type)) {
-                continue;
-            }
-
-            njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
-
-            ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
-
-            if (ext_prop == NULL && prop->type != NJS_WHITEOUT
-                && (prop->enumerable || all))
-            {
-                v = (prop->type != NJS_ACCESSOR)
-                            ? njs_prop_value(prop)
-                            : njs_value_arg(&njs_value_undefined);
-                ret = njs_array_add(vm, items, v);
-                if (njs_slow_path(ret != NJS_OK)) {
-                    return NJS_ERROR;
-                }
-            }
-        }
-
-        njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
-        hash = &object->shared_hash;
-
-        for ( ;; ) {
-            prop = njs_lvlhsh_each(hash, &lhe);
-
-            if (prop == NULL) {
-                break;
-            }
-
-            if (!njs_is_enumerable(&prop->name, type)) {
-                continue;
-            }
-
-            njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
-
-            ret = njs_lvlhsh_find(&object->hash, &lhq);
-
-            if (ret != NJS_OK) {
-                ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
-
-                if (ext_prop == NULL && (prop->enumerable || all)) {
-                    v = (prop->type != NJS_ACCESSOR)
-                                ? njs_prop_value(prop)
-                                : njs_value_arg(&njs_value_undefined);
-                    ret = njs_array_add(vm, items, v);
-                    if (njs_slow_path(ret != NJS_OK)) {
-                        return NJS_ERROR;
-                    }
-                }
-            }
-        }
-
-        break;
-
     case NJS_ENUM_BOTH:
-        for ( ;; ) {
-            prop = njs_lvlhsh_each(hash, &lhe);
+        items_sorted = njs_array_alloc(vm, 1, 0, NJS_ARRAY_SPARE);
+        if (njs_slow_path(items_sorted == NULL)) {
+            return NJS_ERROR;
+        }
 
-            if (prop == NULL) {
-                break;
-            }
+        ret = njs_get_own_ordered_keys(vm, object, parent, items_sorted, kind,
+                                       type, all);
+        if (ret != NJS_OK) {
+            return NJS_ERROR;
+        }
 
-            if (!njs_is_enumerable(&prop->name, type)) {
-                continue;
-            }
+        for (i = 0; i< items_sorted->length; i++) {
 
-            njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
+            lhe.key_hash = 0;
+            njs_object_property_key_set(&lhq, &items_sorted->start[i],
+                                        lhe.key_hash);
 
-            ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
-
-            if (ext_prop == NULL && prop->type != NJS_WHITEOUT
-                && (prop->enumerable || all))
-            {
-                entry = njs_array_alloc(vm, 0, 2, 0);
-                if (njs_slow_path(entry == NULL)) {
+            ret = njs_add_obj_prop_kind(vm, object, &object->hash, &lhq, kind,
+                                        items);
+            if (ret != NJS_DECLINED) {
+                if (ret != NJS_OK) {
                     return NJS_ERROR;
                 }
 
-                njs_string_copy(&entry->start[0], &prop->name);
-
-                v = (prop->type != NJS_ACCESSOR)
-                            ? njs_prop_value(prop)
-                            : njs_value_arg(&njs_value_undefined);
-
-                njs_value_assign(&entry->start[1], v);
-
-                njs_set_array(&value, entry);
-
-                ret = njs_array_add(vm, items, &value);
-                if (njs_slow_path(ret != NJS_OK)) {
+            } else {
+                ret = njs_add_obj_prop_kind(vm, object, &object->shared_hash,
+                                            &lhq, kind, items);
+                njs_assert(ret != NJS_DECLINED);
+                if (ret != NJS_OK) {
                     return NJS_ERROR;
                 }
             }
         }
 
-        njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
-        hash = &object->shared_hash;
-
-        for ( ;; ) {
-            prop = njs_lvlhsh_each(hash, &lhe);
-
-            if (prop == NULL) {
-                break;
-            }
-
-            if (!njs_is_enumerable(&prop->name, type)) {
-                continue;
-            }
-
-            njs_object_property_key_set(&lhq, &prop->name, lhe.key_hash);
-
-            ret = njs_lvlhsh_find(&object->hash, &lhq);
-
-            if (ret != NJS_OK && (prop->enumerable || all)) {
-                ext_prop = njs_object_exist_in_proto(parent, object, &lhq);
-
-                if (ext_prop == NULL) {
-                    entry = njs_array_alloc(vm, 0, 2, 0);
-                    if (njs_slow_path(entry == NULL)) {
-                        return NJS_ERROR;
-                    }
-
-                    njs_string_copy(&entry->start[0], &prop->name);
-                    njs_value_assign(&entry->start[1], njs_prop_value(prop));
-
-                    njs_set_array(&value, entry);
-
-                    ret = njs_array_add(vm, items, &value);
-                    if (njs_slow_path(ret != NJS_OK)) {
-                        return NJS_ERROR;
-                    }
-                }
-            }
-        }
+        njs_array_destroy(vm, items_sorted);
 
         break;
+
     }
 
     return NJS_OK;
@@ -1259,7 +1377,7 @@ done:
 
 static njs_int_t
 njs_object_define_property(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t    ret;
     njs_value_t  *value, *name, *desc, lvalue;
@@ -1285,7 +1403,7 @@ njs_object_define_property(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    vm->retval = *value;
+    njs_value_assign(retval, value);
 
     return NJS_OK;
 }
@@ -1293,7 +1411,7 @@ njs_object_define_property(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_define_properties(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     uint32_t              i, length;
     njs_int_t             ret;
@@ -1350,7 +1468,7 @@ njs_object_define_properties(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     }
 
     ret = NJS_OK;
-    vm->retval = *value;
+    njs_value_assign(retval, value);
 
 done:
 
@@ -1362,7 +1480,7 @@ done:
 
 static njs_int_t
 njs_object_get_own_property_descriptor(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_value_t  lvalue, *value, *property;
 
@@ -1376,13 +1494,13 @@ njs_object_get_own_property_descriptor(njs_vm_t *vm, njs_value_t *args,
 
     property = njs_lvalue_arg(&lvalue, args, nargs, 2);
 
-    return njs_object_prop_descriptor(vm, &vm->retval, value, property);
+    return njs_object_prop_descriptor(vm, retval, value, property);
 }
 
 
 static njs_int_t
 njs_object_get_own_property_descriptors(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t           ret;
     uint32_t            i, length;
@@ -1444,7 +1562,7 @@ njs_object_get_own_property_descriptors(njs_vm_t *vm, njs_value_t *args,
     }
 
     ret = NJS_OK;
-    njs_set_object(&vm->retval, descriptors);
+    njs_set_object(retval, descriptors);
 
 done:
 
@@ -1456,7 +1574,7 @@ done:
 
 static njs_int_t
 njs_object_get_own_property(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t type)
+    njs_uint_t nargs, njs_index_t type, njs_value_t *retval)
 {
     njs_array_t  *names;
     njs_value_t  *value;
@@ -1476,7 +1594,7 @@ njs_object_get_own_property(njs_vm_t *vm, njs_value_t *args,
         return NJS_ERROR;
     }
 
-    njs_set_array(&vm->retval, names);
+    njs_set_array(retval, names);
 
     return NJS_OK;
 }
@@ -1484,7 +1602,7 @@ njs_object_get_own_property(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_object_get_prototype_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     uint32_t     index;
     njs_value_t  *value;
@@ -1492,7 +1610,7 @@ njs_object_get_prototype_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     value = njs_arg(args, nargs, 1);
 
     if (njs_is_object(value)) {
-        njs_object_prototype_proto(vm, NULL, value, NULL, &vm->retval);
+        njs_object_prototype_proto(vm, NULL, value, NULL, retval);
         return NJS_OK;
     }
 
@@ -1500,10 +1618,10 @@ njs_object_get_prototype_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         index = njs_primitive_prototype_index(value->type);
 
         if (njs_is_symbol(value)) {
-            njs_set_object(&vm->retval, &vm->prototypes[index].object);
+            njs_set_object(retval, njs_vm_proto(vm, index));
 
         } else {
-            njs_set_object_value(&vm->retval,
+            njs_set_object_value(retval,
                                  &vm->prototypes[index].object_value);
         }
 
@@ -1519,7 +1637,7 @@ njs_object_get_prototype_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_set_prototype_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t    ret;
     njs_value_t  *value, *proto;
@@ -1539,15 +1657,13 @@ njs_object_set_prototype_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     }
 
     if (njs_slow_path(!njs_is_object(value))) {
-        vm->retval = *value;
-
+        njs_value_assign(retval, value);
         return NJS_OK;
     }
 
     ret = njs_object_set_prototype(vm, njs_object(value), proto);
     if (njs_fast_path(ret == NJS_OK)) {
-        vm->retval = *value;
-
+        njs_value_assign(retval, value);
         return NJS_OK;
     }
 
@@ -1567,7 +1683,7 @@ njs_object_set_prototype_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_set_integrity_level(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t level)
+    njs_uint_t nargs, njs_index_t level, njs_value_t *retval)
 {
     uint32_t           length;
     njs_int_t          ret;
@@ -1581,7 +1697,7 @@ njs_object_set_integrity_level(njs_vm_t *vm, njs_value_t *args,
     value = njs_arg(args, nargs, 1);
 
     if (njs_slow_path(!njs_is_object(value))) {
-        vm->retval = *value;
+        njs_value_assign(retval, value);
         return NJS_OK;
     }
 
@@ -1631,7 +1747,7 @@ njs_object_set_integrity_level(njs_vm_t *vm, njs_value_t *args,
         prop->configurable = 0;
     }
 
-    vm->retval = *value;
+    njs_value_assign(retval, value);
 
     return NJS_OK;
 }
@@ -1639,23 +1755,22 @@ njs_object_set_integrity_level(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_object_test_integrity_level(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t level)
+    njs_uint_t nargs, njs_index_t level, njs_value_t *retval)
 {
     njs_value_t        *value;
     njs_lvlhsh_t       *hash;
     njs_object_t       *object;
     njs_object_prop_t  *prop;
     njs_lvlhsh_each_t  lhe;
-    const njs_value_t  *retval;
 
     value = njs_arg(args, nargs, 1);
 
     if (njs_slow_path(!njs_is_object(value))) {
-        vm->retval = njs_value_true;
+        njs_set_boolean(retval, 1);
         return NJS_OK;
     }
 
-    retval = &njs_value_false;
+    njs_set_boolean(retval, 0);
 
     object = njs_object(value);
 
@@ -1692,32 +1807,30 @@ njs_object_test_integrity_level(njs_vm_t *vm, njs_value_t *args,
         }
     }
 
-    retval = &njs_value_true;
+    njs_set_boolean(retval, 1);
 
 done:
 
-    vm->retval = *retval;
-
     return NJS_OK;
 }
 
 
 static njs_int_t
 njs_object_prevent_extensions(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_value_t  *value;
 
     value = njs_arg(args, nargs, 1);
 
     if (!njs_is_object(value)) {
-        vm->retval = *value;
+        njs_value_assign(retval, value);
         return NJS_OK;
     }
 
     njs_object(&args[1])->extensible = 0;
 
-    vm->retval = *value;
+    njs_value_assign(retval, value);
 
     return NJS_OK;
 }
@@ -1725,22 +1838,18 @@ njs_object_prevent_extensions(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_is_extensible(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
-    njs_value_t        *value;
-    const njs_value_t  *retval;
+    njs_value_t  *value;
 
     value = njs_arg(args, nargs, 1);
 
     if (!njs_is_object(value)) {
-        vm->retval = njs_value_false;
+        njs_set_boolean(retval, 0);
         return NJS_OK;
     }
 
-    retval = njs_object(value)->extensible ? &njs_value_true
-                                           : &njs_value_false;
-
-    vm->retval = *retval;
+    njs_set_boolean(retval, njs_object(value)->extensible);
 
     return NJS_OK;
 }
@@ -1748,7 +1857,7 @@ njs_object_is_extensible(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_object_assign(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     uint32_t              i, j, length;
     njs_int_t             ret;
@@ -1806,7 +1915,7 @@ njs_object_assign(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         njs_array_destroy(vm, names);
     }
 
-    vm->retval = *value;
+    njs_value_assign(retval, value);
 
     return NJS_OK;
 
@@ -1820,9 +1929,9 @@ exception:
 
 static njs_int_t
 njs_object_is(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
-    njs_set_boolean(&vm->retval, njs_values_same(njs_arg(args, nargs, 1),
+    njs_set_boolean(retval, njs_values_same(njs_arg(args, nargs, 1),
                                                  njs_arg(args, nargs, 2)));
 
     return NJS_OK;
@@ -1851,7 +1960,7 @@ njs_primitive_prototype_get_proto(njs_vm_t *vm, njs_object_prop_t *prop,
 
     } else {
         index = njs_primitive_prototype_index(value->type);
-        proto = &vm->prototypes[index].object;
+        proto = njs_vm_proto(vm, index);
     }
 
     if (proto != NULL) {
@@ -1883,7 +1992,7 @@ njs_object_prototype_create(njs_vm_t *vm, njs_object_prop_t *prop,
     function = njs_function(value);
     index = function - vm->constructors;
 
-    if (index >= 0 && index < NJS_OBJ_TYPE_MAX) {
+    if (index >= 0 && (size_t) index < vm->constructors_size) {
         proto = njs_property_prototype_create(vm, &function->object.hash,
                                               &vm->prototypes[index].object);
     }
@@ -1938,10 +2047,10 @@ njs_property_prototype_create(njs_vm_t *vm, njs_lvlhsh_t *hash,
 
 static const njs_object_prop_t  njs_object_constructor_properties[] =
 {
-    NJS_DECLARE_PROP_NAME("Object"),
-
     NJS_DECLARE_PROP_LENGTH(1),
 
+    NJS_DECLARE_PROP_NAME("Object"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_NATIVE("create", njs_object_create, 2, 0),
@@ -2119,7 +2228,7 @@ njs_object_prototype_create_constructor(njs_vm_t *vm, njs_object_prop_t *prop,
             prototype = (njs_object_prototype_t *) object;
             index = prototype - vm->prototypes;
 
-            if (index >= 0 && index < NJS_OBJ_TYPE_MAX) {
+            if (index >= 0 && (size_t) index < vm->constructors_size) {
                 goto found;
             }
 
@@ -2136,7 +2245,7 @@ njs_object_prototype_create_constructor(njs_vm_t *vm, njs_object_prop_t *prop,
 
 found:
 
-    njs_set_function(&constructor, &vm->constructors[index]);
+    njs_set_function(&constructor, &njs_vm_ctor(vm, index));
     setval = &constructor;
 
     cons = njs_property_constructor_set(vm, &prototype->object.hash, setval);
@@ -2190,12 +2299,12 @@ njs_property_constructor_set(njs_vm_t *vm, njs_lvlhsh_t *hash,
 
 static njs_int_t
 njs_object_prototype_value_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
-    vm->retval = *njs_argument(args, 0);
+    njs_value_assign(retval, njs_argument(args, 0));
 
-    if (!njs_is_object(&vm->retval)) {
-        return njs_value_to_object(vm, &vm->retval);
+    if (!njs_is_object(retval)) {
+        return njs_value_to_object(vm, retval);
     }
 
     return NJS_OK;
@@ -2228,9 +2337,9 @@ static const njs_value_t  njs_object_arguments_string =
 
 njs_int_t
 njs_object_prototype_to_string(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
-    return njs_object_to_string(vm, &args[0], &vm->retval);
+    return njs_object_to_string(vm, &args[0], retval);
 }
 
 
@@ -2334,7 +2443,7 @@ njs_object_to_string(njs_vm_t *vm, njs_value_t *this, njs_value_t *retval)
 
 static njs_int_t
 njs_object_prototype_has_own_property(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t             ret;
     njs_value_t           *value, *property, lvalue;
@@ -2363,11 +2472,11 @@ njs_object_prototype_has_own_property(njs_vm_t *vm, njs_value_t *args,
 
     switch (ret) {
     case NJS_OK:
-        vm->retval = njs_value_true;
+        njs_set_boolean(retval, 1);
         return NJS_OK;
 
     case NJS_DECLINED:
-        vm->retval = njs_value_false;
+        njs_set_boolean(retval, 0);
         return NJS_OK;
 
     case NJS_ERROR:
@@ -2379,11 +2488,10 @@ njs_object_prototype_has_own_property(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_object_prototype_prop_is_enumerable(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t             ret;
     njs_value_t           *value, *property, lvalue;
-    const njs_value_t     *retval;
     njs_object_prop_t     *prop;
     njs_property_query_t  pq;
 
@@ -2411,11 +2519,11 @@ njs_object_prototype_prop_is_enumerable(njs_vm_t *vm, njs_value_t *args,
     switch (ret) {
     case NJS_OK:
         prop = pq.lhq.value;
-        retval = prop->enumerable ? &njs_value_true : &njs_value_false;
+        njs_set_boolean(retval, prop->enumerable);
         break;
 
     case NJS_DECLINED:
-        retval = &njs_value_false;
+        njs_set_boolean(retval, 0);
         break;
 
     case NJS_ERROR:
@@ -2423,26 +2531,22 @@ njs_object_prototype_prop_is_enumerable(njs_vm_t *vm, njs_value_t *args,
         return NJS_ERROR;
     }
 
-    vm->retval = *retval;
-
     return NJS_OK;
 }
 
 
 static njs_int_t
 njs_object_prototype_is_prototype_of(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
-    njs_value_t        *prototype, *value;
-    njs_object_t       *object, *proto;
-    const njs_value_t  *retval;
+    njs_value_t   *prototype, *value;
+    njs_object_t  *object, *proto;
 
     if (njs_slow_path(njs_is_null_or_undefined(njs_argument(args, 0)))) {
         njs_type_error(vm, "cannot convert undefined to object");
         return NJS_ERROR;
     }
 
-    retval = &njs_value_false;
     prototype = &args[0];
     value = njs_arg(args, nargs, 1);
 
@@ -2454,14 +2558,14 @@ njs_object_prototype_is_prototype_of(njs_vm_t *vm, njs_value_t *args,
             object = object->__proto__;
 
             if (object == proto) {
-                retval = &njs_value_true;
-                break;
+                njs_set_boolean(retval, 1);
+                return NJS_OK;
             }
 
         } while (object != NULL);
     }
 
-    vm->retval = *retval;
+    njs_set_boolean(retval, 0);
 
     return NJS_OK;
 }
