@@ -68,6 +68,7 @@ njs_object_prop_alloc2(njs_vm_t *vm, const njs_value_t *name,
     }
 
     prop->type = type;
+    prop->enum_in_object_hash = 0;
 
     if (flags != NJS_OBJECT_PROP_UNSET) {
         prop->enumerable = !!(flags & NJS_OBJECT_PROP_ENUMERABLE);
@@ -189,6 +190,7 @@ njs_object_prop_define(njs_vm_t *vm, njs_value_t *object,
     uint32_t              length, index;
     njs_int_t             ret;
     njs_array_t           *array;
+    njs_value_t           retval;
     njs_object_prop_t     *prop, *prev;
     njs_property_query_t  pq;
 
@@ -557,7 +559,7 @@ done:
         if (prev->type == NJS_PROPERTY_HANDLER) {
             if (prev->writable) {
                 ret = njs_prop_handler(prev)(vm, prev, object,
-                                             njs_prop_value(prop), &vm->retval);
+                                             njs_prop_value(prop), &retval);
                 if (njs_slow_path(ret == NJS_ERROR)) {
                     return ret;
                 }
