@@ -18,13 +18,16 @@ struct njs_regexp_group_s {
 static void *njs_regexp_malloc(size_t size, void *memory_data);
 static void njs_regexp_free(void *p, void *memory_data);
 static njs_int_t njs_regexp_prototype_source(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused);
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
 static int njs_regexp_pattern_compile(njs_vm_t *vm, njs_regex_t *regex,
     u_char *source, size_t len, njs_regex_flags_t flags);
 static u_char *njs_regexp_compile_trace_handler(njs_trace_t *trace,
     njs_trace_data_t *td, u_char *start);
 static u_char *njs_regexp_match_trace_handler(njs_trace_t *trace,
     njs_trace_data_t *td, u_char *start);
+#define NJS_REGEXP_FLAG_TEST           1
+static njs_int_t njs_regexp_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s,
+    unsigned flags, njs_value_t *retval);
 static njs_array_t *njs_regexp_exec_result(njs_vm_t *vm, njs_value_t *r,
     njs_utf8_t utf8, njs_string_prop_t *string, njs_regex_match_data_t *data);
 static njs_int_t njs_regexp_string_create(njs_vm_t *vm, njs_value_t *value,
@@ -107,7 +110,7 @@ njs_regexp_value_flags(njs_vm_t *vm, const njs_value_t *regexp)
 
 static njs_int_t
 njs_regexp_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     u_char              *start;
     njs_int_t           ret;
@@ -118,13 +121,11 @@ njs_regexp_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     pattern = njs_arg(args, nargs, 1);
 
     if (njs_is_regexp(pattern)) {
-        ret = njs_regexp_prototype_source(vm, pattern, 1, 0);
+        ret = njs_regexp_prototype_source(vm, pattern, 1, 0, &source);
         if (njs_slow_path(ret != NJS_OK)) {
             return ret;
         }
 
-        source = vm->retval;
-
         re_flags = njs_regexp_value_flags(vm, pattern);
 
         pattern = &source;
@@ -166,7 +167,7 @@ njs_regexp_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
     njs_string_get(pattern, &string);
 
-    return njs_regexp_create(vm, &vm->retval, string.start, string.length,
+    return njs_regexp_create(vm, retval, string.start, string.length,
                              re_flags);
 }
 
@@ -265,9 +266,10 @@ njs_regexp_pattern_create(njs_vm_t *vm, u_char *start, size_t length,
     njs_regex_flags_t flags)
 {
     int                   ret;
-    u_char                *p;
+    u_char                *p, *end;
     size_t                size;
     njs_str_t             text;
+    njs_bool_t            in;
     njs_uint_t            n;
     njs_regex_t           *regex;
     njs_regexp_group_t    *group;
@@ -276,6 +278,42 @@ njs_regexp_pattern_create(njs_vm_t *vm, u_char *start, size_t length,
     text.start = start;
     text.length = length;
 
+    in = 0;
+    end = start + length;
+
+    for (p = start; p < end; p++) {
+
+        switch (*p) {
+        case '[':
+            in = 1;
+            break;
+
+        case ']':
+            in = 0;
+            break;
+
+        case '\\':
+            p++;
+            break;
+
+        case '+':
+            if (njs_slow_path(!in
+                              && (p - 1 > start)
+                              && (p[-1] == '+'|| p[-1] == '*' || p[-1] == '?'))
+                              && (p - 2 >= start && p[-2] != '\\'))
+            {
+                /**
+                 * PCRE possessive quantifiers `++`, `*+`, `?+`
+                 * are not allowed in JavaScript. Whereas `[++]` or `\?+` are
+                 * allowed.
+                 */
+                goto nothing_to_repeat;
+            }
+
+            break;
+        }
+    }
+
     ret = njs_regex_escape(vm->mem_pool, &text);
     if (njs_slow_path(ret != NJS_OK)) {
         njs_memory_error(vm);
@@ -312,6 +350,8 @@ njs_regexp_pattern_create(njs_vm_t *vm, u_char *start, size_t length,
         goto fail;
     }
 
+    njs_set_invalid(&vm->exception);
+
     ret = njs_regexp_pattern_compile(vm, &pattern->regex[1],
                                   &pattern->source[0], text.length,
                                   flags | NJS_REGEX_UTF8);
@@ -364,14 +404,18 @@ njs_regexp_pattern_create(njs_vm_t *vm, u_char *start, size_t length,
         } while (n != pattern->ngroups);
     }
 
-    njs_set_undefined(&vm->retval);
-
     return pattern;
 
 fail:
 
     njs_mp_free(vm->mem_pool, pattern);
     return NULL;
+
+nothing_to_repeat:
+
+    njs_syntax_error(vm, "Invalid regular expression \"%V\" nothing to repeat",
+                     &text);
+    return NULL;
 }
 
 
@@ -462,7 +506,7 @@ njs_regexp_alloc(njs_vm_t *vm, njs_regexp_pattern_t *pattern)
     if (njs_fast_path(regexp != NULL)) {
         njs_lvlhsh_init(&regexp->object.hash);
         regexp->object.shared_hash = vm->shared->regexp_instance_hash;
-        regexp->object.__proto__ = &vm->prototypes[NJS_OBJ_TYPE_REGEXP].object;
+        regexp->object.__proto__ = njs_vm_proto(vm, NJS_OBJ_TYPE_REGEXP);
         regexp->object.slots = NULL;
         regexp->object.type = NJS_REGEXP;
         regexp->object.shared = 0;
@@ -496,19 +540,19 @@ njs_regexp_prototype_last_index(njs_vm_t *vm, njs_object_prop_t *unused,
 
     if (setval != NULL) {
         regexp->last_index = *setval;
-        *retval  = *setval;
+        njs_value_assign(retval, setval);
 
         return NJS_OK;
     }
 
-    *retval = regexp->last_index;
+    njs_value_assign(retval, &regexp->last_index);
     return NJS_OK;
 }
 
 
 static njs_int_t
 njs_regexp_prototype_flags(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     u_char       *p;
     njs_int_t    ret;
@@ -568,13 +612,13 @@ njs_regexp_prototype_flags(njs_vm_t *vm, njs_value_t *args,
         *p++ = 'y';
     }
 
-    return njs_string_new(vm, &vm->retval, dst, p - dst, p - dst);
+    return njs_string_new(vm, retval, dst, p - dst, p - dst);
 }
 
 
 static njs_int_t
 njs_regexp_prototype_flag(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t flag)
+    njs_uint_t nargs, njs_index_t flag, njs_value_t *retval)
 {
     unsigned              yn;
     njs_value_t           *this;
@@ -587,8 +631,8 @@ njs_regexp_prototype_flag(njs_vm_t *vm, njs_value_t *args,
     }
 
     if (njs_slow_path(!njs_is_regexp(this))) {
-        if (njs_object(this) == &vm->prototypes[NJS_OBJ_TYPE_REGEXP].object) {
-            njs_set_undefined(&vm->retval);
+        if (njs_object(this) == njs_vm_proto(vm, NJS_OBJ_TYPE_REGEXP)) {
+            njs_set_undefined(retval);
             return NJS_OK;
         }
 
@@ -617,7 +661,7 @@ njs_regexp_prototype_flag(njs_vm_t *vm, njs_value_t *args,
         break;
     }
 
-    njs_set_boolean(&vm->retval, yn);
+    njs_set_boolean(retval, yn);
 
     return NJS_OK;
 }
@@ -625,7 +669,7 @@ njs_regexp_prototype_flag(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_regexp_prototype_source(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_str_t             src;
     njs_value_t           *this;
@@ -638,8 +682,8 @@ njs_regexp_prototype_source(njs_vm_t *vm, njs_value_t *args,
     }
 
     if (njs_slow_path(!njs_is_regexp(this))) {
-        if (njs_object(this) == &vm->prototypes[NJS_OBJ_TYPE_REGEXP].object) {
-            vm->retval = njs_string_empty_regexp;
+        if (njs_object(this) == njs_vm_proto(vm, NJS_OBJ_TYPE_REGEXP)) {
+            njs_value_assign(retval, &njs_string_empty_regexp);
             return NJS_OK;
         }
 
@@ -652,13 +696,13 @@ njs_regexp_prototype_source(njs_vm_t *vm, njs_value_t *args,
     src.start = pattern->source;
     src.length = njs_strlen(pattern->source);
 
-    return njs_string_decode_utf8(vm, &vm->retval, &src);
+    return njs_string_decode_utf8(vm, retval, &src);
 }
 
 
 static njs_int_t
 njs_regexp_prototype_to_string(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     u_char             *p;
     size_t             size, length;
@@ -710,7 +754,7 @@ njs_regexp_prototype_to_string(njs_vm_t *vm, njs_value_t *args,
         length = 0;
     }
 
-    p = njs_string_alloc(vm, &vm->retval, size, length);
+    p = njs_string_alloc(vm, retval, size, length);
     if (njs_slow_path(p == NULL)) {
         return NJS_ERROR;
     }
@@ -789,10 +833,10 @@ njs_regexp_to_string(njs_vm_t *vm, njs_value_t *retval,
 
 static njs_int_t
 njs_regexp_prototype_test(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t    ret;
-    njs_value_t  *r, *string, lvalue, retval;
+    njs_value_t  *r, *string, lvalue, value;
 
     r = njs_argument(args, 0);
 
@@ -808,12 +852,12 @@ njs_regexp_prototype_test(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    ret = njs_regexp_exec(vm, r, string, &retval);
+    ret = njs_regexp_exec(vm, r, string, NJS_REGEXP_FLAG_TEST, &value);
     if (njs_slow_path(ret != NJS_OK)) {
         return NJS_ERROR;
     }
 
-    njs_set_boolean(&vm->retval, !njs_is_null(&retval));
+    njs_set_boolean(retval, !njs_is_null(&value));
 
     return NJS_OK;
 }
@@ -824,10 +868,11 @@ njs_regexp_prototype_test(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
  */
 static njs_int_t
 njs_regexp_builtin_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s,
-    njs_value_t *retval)
+    unsigned flags, njs_value_t *retval)
 {
-    size_t                  length, offset;
+    size_t                  c, length, offset;
     int64_t                 last_index;
+    uint32_t                index;
     njs_int_t               ret;
     njs_utf8_t              utf8;
     njs_value_t             value;
@@ -891,15 +936,49 @@ njs_regexp_builtin_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s,
         offset = last_index;
 
     } else {
-        offset = njs_string_utf8_offset(string.start,
-                                        string.start + string.size, last_index)
-                 - string.start;
+        if ((size_t) last_index < string.length) {
+            offset = njs_string_utf8_offset(string.start,
+                                            string.start + string.size,
+                                            last_index)
+                     - string.start;
+        } else {
+            offset = string.size;
+        }
     }
 
     ret = njs_regexp_match(vm, &pattern->regex[type], string.start, offset,
                            string.size, match_data);
     if (ret >= 0) {
+        if (pattern->global || pattern->sticky) {
+            c = njs_regex_capture(match_data, 1);
+
+            if (utf8 == NJS_STRING_UTF8) {
+                index = njs_string_index(&string, c);
+
+            } else {
+                index = c;
+            }
+
+            njs_set_number(&value, index);
+            ret = njs_value_property_set(vm, r,
+                                         njs_value_arg(&njs_string_lindex),
+                                         &value);
+            if (njs_slow_path(ret != NJS_OK)) {
+                njs_regex_match_data_free(match_data, vm->regex_generic_ctx);
+                return NJS_ERROR;
+            }
+        }
+
+        if (flags & NJS_REGEXP_FLAG_TEST) {
+            njs_regex_match_data_free(match_data, vm->regex_generic_ctx);
+            njs_set_boolean(retval, 1);
+            return NJS_OK;
+        }
+
         result = njs_regexp_exec_result(vm, r, utf8, &string, match_data);
+
+        njs_regex_match_data_free(match_data, vm->regex_generic_ctx);
+
         if (njs_slow_path(result == NULL)) {
             return NJS_ERROR;
         }
@@ -908,8 +987,9 @@ njs_regexp_builtin_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s,
         return NJS_OK;
     }
 
+    njs_regex_match_data_free(match_data, vm->regex_generic_ctx);
+
     if (njs_slow_path(ret == NJS_ERROR)) {
-        njs_regex_match_data_free(match_data, vm->regex_generic_ctx);
         return NJS_ERROR;
     }
 
@@ -930,6 +1010,12 @@ not_found:
 }
 
 
+static njs_exotic_slots_t njs_regexp_prototype_exotic_slots = {
+    .prop_handler = NULL,
+    .keys = NULL,
+};
+
+
 static njs_array_t *
 njs_regexp_exec_result(njs_vm_t *vm, njs_value_t *r, njs_utf8_t utf8,
     njs_string_prop_t *string, njs_regex_match_data_t *match_data)
@@ -941,7 +1027,7 @@ njs_regexp_exec_result(njs_vm_t *vm, njs_value_t *r, njs_utf8_t utf8,
     njs_int_t             ret;
     njs_uint_t            i, n;
     njs_array_t           *array;
-    njs_value_t           name, value;
+    njs_value_t           name;
     njs_object_t          *groups;
     njs_regexp_t          *regexp;
     njs_object_prop_t     *prop;
@@ -960,6 +1046,8 @@ njs_regexp_exec_result(njs_vm_t *vm, njs_value_t *r, njs_utf8_t utf8,
         goto fail;
     }
 
+    array->object.slots = &njs_regexp_prototype_exotic_slots;
+
     for (i = 0; i < pattern->ncaptures; i++) {
         n = 2 * i;
         c = njs_regex_capture(match_data, n);
@@ -1003,24 +1091,6 @@ njs_regexp_exec_result(njs_vm_t *vm, njs_value_t *r, njs_utf8_t utf8,
 
     njs_set_number(&prop->u.value, index);
 
-    if (pattern->global || pattern->sticky) {
-        c = njs_regex_capture(match_data, 1);
-
-        if (utf8 == NJS_STRING_UTF8) {
-            index = njs_string_index(string, c);
-
-        } else {
-            index = c;
-        }
-
-        njs_set_number(&value, index);
-        ret = njs_value_property_set(vm, r, njs_value_arg(&njs_string_lindex),
-                                     &value);
-        if (njs_slow_path(ret != NJS_OK)) {
-            goto fail;
-        }
-    }
-
     lhq.key_hash = NJS_INDEX_HASH;
     lhq.key = njs_str_value("index");
     lhq.replace = 0;
@@ -1113,15 +1183,57 @@ fail:
 
 done:
 
-    njs_regex_match_data_free(match_data, vm->regex_generic_ctx);
-
     return (ret == NJS_OK) ? array : NULL;
 }
 
 
+static void
+njs_regexp_exec_result_free(njs_vm_t *vm, njs_array_t *result)
+{
+    njs_uint_t          n;
+    njs_value_t         *start;
+    njs_flathsh_t       *hash;
+    njs_object_prop_t   *prop;
+    njs_lvlhsh_each_t   lhe;
+    njs_lvlhsh_query_t  lhq;
+
+    if (result->object.fast_array) {
+        start = result->start;
+
+        for (n = 0; n < result->length; n++) {
+            if (start[n].short_string.size == NJS_STRING_LONG) {
+                njs_mp_free(vm->mem_pool, start[n].long_string.data);
+            }
+        }
+    }
+
+    njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
+
+    hash = &result->object.hash;
+
+    for ( ;; ) {
+        prop = njs_flathsh_each(hash, &lhe);
+
+        if (prop == NULL) {
+            break;
+        }
+
+        njs_mp_free(vm->mem_pool, prop);
+    }
+
+
+    lhq.pool = vm->mem_pool;
+    lhq.proto = &njs_object_hash_proto;
+
+    njs_flathsh_destroy(hash, &lhq);
+
+    njs_array_destroy(vm, result);
+}
+
+
 njs_int_t
 njs_regexp_prototype_exec(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     njs_int_t    ret;
     njs_value_t  *r, *s;
@@ -1141,16 +1253,18 @@ njs_regexp_prototype_exec(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return ret;
     }
 
-    return njs_regexp_builtin_exec(vm, r, s, &vm->retval);
+    return njs_regexp_builtin_exec(vm, r, s,
+                                   njs_number(njs_arg(args, nargs, 2)), retval);
 }
 
 
-njs_int_t
-njs_regexp_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s,
+static njs_int_t
+njs_regexp_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s, unsigned flags,
     njs_value_t *retval)
 {
     njs_int_t    ret;
     njs_value_t  exec;
+    njs_value_t  arguments[2];
 
     static const njs_value_t  string_exec = njs_string("exec");
 
@@ -1160,12 +1274,30 @@ njs_regexp_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s,
     }
 
     if (njs_is_function(&exec)) {
-        ret = njs_function_call(vm, njs_function(&exec), r, s, 1, retval);
+        njs_value_assign(&arguments[0], s);
+
+        if (flags) {
+            njs_set_number(&arguments[1], flags);
+        }
+
+        ret = njs_function_call(vm, njs_function(&exec), r, arguments,
+                                flags ? 2 : 1, retval);
         if (njs_slow_path(ret == NJS_ERROR)) {
             return NJS_ERROR;
         }
 
-        if (njs_slow_path(!njs_is_object(retval) && !njs_is_null(retval))) {
+        if (njs_slow_path(!njs_is_null(retval)
+                          && (flags & NJS_REGEXP_FLAG_TEST
+                              && !njs_is_boolean(retval))))
+        {
+            njs_type_error(vm, "unexpected \"%s\" retval in njs_regexp_exec()",
+                           njs_type_string(retval->type));
+            return NJS_ERROR;
+        }
+
+        if (njs_slow_path(!njs_is_null(retval)
+                          && (!flags && !njs_is_object(retval))))
+        {
             njs_type_error(vm, "unexpected \"%s\" retval in njs_regexp_exec()",
                            njs_type_string(retval->type));
             return NJS_ERROR;
@@ -1179,7 +1311,7 @@ njs_regexp_exec(njs_vm_t *vm, njs_value_t *r, njs_value_t *s,
         return NJS_ERROR;
     }
 
-    return njs_regexp_builtin_exec(vm, r, s, retval);
+    return njs_regexp_builtin_exec(vm, r, s, flags, retval);
 }
 
 
@@ -1195,7 +1327,7 @@ njs_regexp_string_create(njs_vm_t *vm, njs_value_t *value, u_char *start,
 
 njs_int_t
 njs_regexp_prototype_symbol_replace(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     int64_t            n, last_index, ncaptures, pos, next_pos, length;
     njs_str_t          rep, m;
@@ -1206,7 +1338,7 @@ njs_regexp_prototype_symbol_replace(njs_vm_t *vm, njs_value_t *args,
     njs_bool_t         global;
     njs_array_t        *array;
     njs_value_t        *arguments, *r, *rx, *string, *replace;
-    njs_value_t        s_lvalue, r_lvalue, value, matched, groups, retval;
+    njs_value_t        s_lvalue, r_lvalue, value, matched, groups;
     njs_function_t     *func_replace;
     njs_string_prop_t  s;
 
@@ -1276,7 +1408,7 @@ njs_regexp_prototype_symbol_replace(njs_vm_t *vm, njs_value_t *args,
             goto exception;
         }
 
-        ret = njs_regexp_exec(vm, rx, string, r);
+        ret = njs_regexp_exec(vm, rx, string, 0, r);
         if (njs_slow_path(ret != NJS_OK)) {
             goto exception;
         }
@@ -1412,7 +1544,7 @@ njs_regexp_prototype_symbol_replace(njs_vm_t *vm, njs_value_t *args,
 
             ret = njs_string_get_substitution(vm, &matched, string, pos,
                                               arguments, ncaptures, &groups,
-                                              replace, &retval);
+                                              replace, retval);
 
         } else {
             ret = njs_array_expand(vm, array, 0,
@@ -1431,14 +1563,16 @@ njs_regexp_prototype_symbol_replace(njs_vm_t *vm, njs_value_t *args,
 
             ret = njs_function_call(vm, func_replace,
                                     njs_value_arg(&njs_value_undefined),
-                                    arguments, n, &retval);
+                                    arguments, n, retval);
         }
 
+        njs_array_destroy(vm, array);
+
         if (njs_slow_path(ret == NJS_ERROR)) {
             return NJS_ERROR;
         }
 
-        ret = njs_value_to_string(vm, &retval, &retval);
+        ret = njs_value_to_string(vm, retval, retval);
         if (njs_slow_path(ret != NJS_OK)) {
             goto exception;
         }
@@ -1446,20 +1580,29 @@ njs_regexp_prototype_symbol_replace(njs_vm_t *vm, njs_value_t *args,
         if (pos >= next_pos) {
             njs_chb_append(&chain, &s.start[next_pos], pos - next_pos);
 
-            njs_string_get(&retval, &rep);
+            njs_string_get(retval, &rep);
             njs_chb_append_str(&chain, &rep);
 
             njs_string_get(&matched, &m);
 
             next_pos = pos + (int64_t) m.length;
         }
+
+        if (!func_replace && njs_object_slots(r)) {
+            /*
+              * Doing free here ONLY for non-function replace, because
+              * otherwise we cannot be certain the result of match
+              * was not stored elsewhere.
+              */
+            njs_regexp_exec_result_free(vm, njs_array(r));
+        }
     }
 
     if (next_pos < (int64_t) s.size) {
         njs_chb_append(&chain, &s.start[next_pos], s.size - next_pos);
     }
 
-    ret = njs_string_create_chb(vm, &vm->retval, &chain);
+    ret = njs_string_create_chb(vm, retval, &chain);
     if (njs_slow_path(ret != NJS_OK)) {
         ret = NJS_ERROR;
         goto exception;
@@ -1478,20 +1621,21 @@ exception:
 
 static njs_int_t
 njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     u_char             *dst;
     int64_t            e, i, p, q, ncaptures, length;
+    ssize_t            len;
     uint32_t           limit;
     njs_int_t          ret;
     njs_bool_t         sticky;
     njs_utf8_t         utf8;
     njs_array_t        *array;
     njs_value_t        *rx, *string, *value;
-    njs_value_t        r, z, this, s_lvalue, retval, setval, constructor;
+    njs_value_t        r, z, this, s_lvalue, setval, constructor;
     njs_object_t       *object;
     const u_char       *start, *end;
-    njs_string_prop_t  s;
+    njs_string_prop_t  s, sv;
     njs_value_t        arguments[2];
 
     static const njs_value_t  string_lindex = njs_string("lastIndex");
@@ -1511,24 +1655,24 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
         return ret;
     }
 
-    njs_set_function(&constructor, &vm->constructors[NJS_OBJ_TYPE_REGEXP]);
+    njs_set_function(&constructor, &njs_vm_ctor(vm, NJS_OBJ_TYPE_REGEXP));
 
     ret = njs_value_species_constructor(vm, rx, &constructor, &constructor);
     if (njs_slow_path(ret != NJS_OK)) {
         return ret;
     }
 
-    ret = njs_value_property(vm, rx, njs_value_arg(&string_flags), &retval);
+    ret = njs_value_property(vm, rx, njs_value_arg(&string_flags), retval);
     if (njs_slow_path(ret == NJS_ERROR)) {
         return NJS_ERROR;
     }
 
-    ret = njs_value_to_string(vm, &retval, &retval);
+    ret = njs_value_to_string(vm, retval, retval);
     if (njs_slow_path(ret != NJS_OK)) {
         return ret;
     }
 
-    (void) njs_string_prop(&s, &retval);
+    (void) njs_string_prop(&s, retval);
 
     sticky = memchr(s.start, 'y', s.size) != NULL;
 
@@ -1553,7 +1697,7 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
         *dst++ = 'y';
 
     } else {
-        arguments[1] = retval;
+        njs_value_assign(&arguments[1], retval);
     }
 
     ret = njs_function_call2(vm, njs_function(&constructor), &this,
@@ -1586,7 +1730,7 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
     length = njs_string_prop(&s, string);
 
     if (njs_slow_path(s.size == 0)) {
-        ret = njs_regexp_exec(vm, rx, string, &z);
+        ret = njs_regexp_exec(vm, rx, string, NJS_REGEXP_FLAG_TEST, &z);
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
@@ -1615,7 +1759,7 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
             return NJS_ERROR;
         }
 
-        ret = njs_regexp_exec(vm, rx, string, &z);
+        ret = njs_regexp_exec(vm, rx, string, 0, &z);
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
@@ -1626,12 +1770,12 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
         }
 
         ret = njs_value_property(vm, rx, njs_value_arg(&string_lindex),
-                                 &retval);
+                                 retval);
         if (njs_slow_path(ret == NJS_ERROR)) {
             return NJS_ERROR;
         }
 
-        ret = njs_value_to_length(vm, &retval, &e);
+        ret = njs_value_to_length(vm, retval, &e);
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
@@ -1639,6 +1783,10 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
         e = njs_min(e, length);
 
         if (e == p) {
+            if (njs_object_slots(&z)) {
+                njs_regexp_exec_result_free(vm, njs_array(&z));
+            }
+
             q = q + 1;
             continue;
         }
@@ -1652,12 +1800,18 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
             end = &s.start[q];
         }
 
-        ret = njs_string_split_part_add(vm, array, utf8, start, end - start);
+        len = njs_string_calc_length(utf8, start, end - start);
+
+        ret = njs_array_string_add(vm, array, start, end - start, len);
         if (njs_slow_path(ret != NJS_OK)) {
             return ret;
         }
 
         if (array->length == limit) {
+            if (njs_object_slots(&z)) {
+                njs_regexp_exec_result_free(vm, njs_array(&z));
+            }
+
             goto done;
         }
 
@@ -1671,21 +1825,32 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
         ncaptures = njs_max(ncaptures - 1, 0);
 
         for (i = 1; i <= ncaptures; i++) {
-            value = njs_array_push(vm, array);
-            if (njs_slow_path(value == NULL)) {
-                return NJS_ERROR;
-            }
-
-            ret = njs_value_property_i64(vm, &z, i, value);
+            ret = njs_value_property_i64(vm, &z, i, retval);
             if (njs_slow_path(ret == NJS_ERROR)) {
                 return NJS_ERROR;
             }
 
+            (void) njs_string_prop(&sv, retval);
+
+            ret = njs_array_string_add(vm, array, sv.start, sv.size,
+                                       sv.length);
+            if (njs_slow_path(ret != NJS_OK)) {
+                return ret;
+            }
+
             if (array->length == limit) {
+                if (njs_object_slots(&z)) {
+                    njs_regexp_exec_result_free(vm, njs_array(&z));
+                }
+
                 goto done;
             }
         }
 
+        if (njs_object_slots(&z)) {
+            njs_regexp_exec_result_free(vm, njs_array(&z));
+        }
+
         q = p;
     }
 
@@ -1700,7 +1865,9 @@ njs_regexp_prototype_symbol_split(njs_vm_t *vm, njs_value_t *args,
         start = &s.start[p];
     }
 
-    ret = njs_string_split_part_add(vm, array, utf8, start, end - start);
+    len = njs_string_calc_length(utf8, start, end - start);
+
+    ret = njs_array_string_add(vm, array, start, end - start, len);
     if (njs_slow_path(ret != NJS_OK)) {
         return ret;
     }
@@ -1718,7 +1885,7 @@ single:
 
 done:
 
-    njs_set_array(&vm->retval, array);
+    njs_set_array(retval, array);
 
     return NJS_OK;
 }
@@ -1726,10 +1893,10 @@ done:
 
 static const njs_object_prop_t  njs_regexp_constructor_properties[] =
 {
-    NJS_DECLARE_PROP_NAME("RegExp"),
-
     NJS_DECLARE_PROP_LENGTH(2),
 
+    NJS_DECLARE_PROP_NAME("RegExp"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 };
 
