@@ -180,7 +180,7 @@ njs_typed_array_alloc(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
     njs_lvlhsh_init(&array->object.hash);
     njs_lvlhsh_init(&array->object.shared_hash);
-    array->object.__proto__ = &vm->prototypes[type].object;
+    array->object.__proto__ = njs_vm_proto(vm, type);
     array->object.type = NJS_TYPED_ARRAY;
     array->object.extensible = 1;
     array->object.fast_array = 1;
@@ -197,7 +197,7 @@ memory_error:
 
 static njs_int_t
 njs_typed_array_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t magic)
+    njs_index_t magic, njs_value_t *retval)
 {
     njs_typed_array_t  *array;
 
@@ -211,7 +211,7 @@ njs_typed_array_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    njs_set_typed_array(&vm->retval, array);
+    njs_set_typed_array(retval, array);
 
     return NJS_OK;
 }
@@ -222,19 +222,9 @@ njs_typed_array_create(njs_vm_t *vm, njs_value_t *constructor,
     njs_value_t *args, njs_uint_t nargs, njs_value_t *retval)
 {
     njs_int_t          ret;
-    njs_value_t        this;
-    njs_object_t       *object;
     njs_typed_array_t  *array;
 
-    object = njs_function_new_object(vm, constructor);
-    if (njs_slow_path(object == NULL)) {
-        return NJS_ERROR;
-    }
-
-    njs_set_object(&this, object);
-
-    ret = njs_function_call2(vm, njs_function(constructor), &this, args,
-                             nargs, retval, 1);
+    ret = njs_value_construct(vm, constructor, args, nargs, retval);
     if (njs_slow_path(ret != NJS_OK)) {
         return NJS_ERROR;
     }
@@ -274,7 +264,7 @@ njs_typed_array_species_create(njs_vm_t *vm, njs_value_t *exemplar,
 
     array = njs_typed_array(exemplar);
 
-    njs_set_function(&constructor, &vm->constructors[array->type]);
+    njs_set_function(&constructor, &njs_vm_ctor(vm, array->type));
 
     ret = njs_value_species_constructor(vm, exemplar, &constructor,
                                         &constructor);
@@ -288,7 +278,7 @@ njs_typed_array_species_create(njs_vm_t *vm, njs_value_t *exemplar,
 
 static njs_int_t
 njs_typed_array_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     double             num;
     uint32_t           length, i;
@@ -308,12 +298,12 @@ njs_typed_array_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     length = nargs - 1;
 
     njs_set_number(&argument, length);
-    ret = njs_typed_array_create(vm, this, &argument, 1, &vm->retval);
+    ret = njs_typed_array_create(vm, this, &argument, 1, retval);
     if (njs_slow_path(ret != NJS_OK)) {
         return NJS_ERROR;
     }
 
-    array = njs_typed_array(&vm->retval);
+    array = njs_typed_array(retval);
 
     for (i = 0; i < length; i++) {
         ret = njs_value_to_number(vm, njs_argument(args, i + 1), &num);
@@ -324,7 +314,7 @@ njs_typed_array_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         njs_typed_array_prop_set(vm, array, i, num);
     }
 
-    njs_set_typed_array(&vm->retval, array);
+    njs_set_typed_array(retval, array);
 
     return NJS_OK;
 }
@@ -332,13 +322,13 @@ njs_typed_array_of(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_typed_array_from(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     double             num;
     int64_t            length, i;
     njs_int_t          ret;
     njs_value_t        *this, *source, *mapfn;
-    njs_value_t        arguments[3], retval;
+    njs_value_t        arguments[3], value;
     njs_function_t     *function;
     njs_typed_array_t  *array;
 
@@ -375,30 +365,30 @@ njs_typed_array_from(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     }
 
     njs_set_number(&arguments[0], length);
-    ret = njs_typed_array_create(vm, this, arguments, 1, &vm->retval);
+    ret = njs_typed_array_create(vm, this, arguments, 1, retval);
     if (njs_slow_path(ret != NJS_OK)) {
         return NJS_ERROR;
     }
 
-    array = njs_typed_array(&vm->retval);
+    array = njs_typed_array(retval);
     arguments[0] = *njs_arg(args, nargs, 3);
 
     for (i = 0; i < length; i++) {
-        ret = njs_value_property_i64(vm, source, i, &retval);
+        ret = njs_value_property_i64(vm, source, i, &value);
         if (njs_slow_path(ret == NJS_ERROR)) {
             return NJS_ERROR;
         }
 
         if (function != NULL) {
-            arguments[1] = retval;
+            njs_value_assign(&arguments[1], &value);
             njs_set_number(&arguments[2], i);
-            ret = njs_function_apply(vm, function, arguments, 3, &retval);
+            ret = njs_function_apply(vm, function, arguments, 3, &value);
             if (njs_slow_path(ret != NJS_OK)) {
                 return NJS_ERROR;
             }
         }
 
-        ret = njs_value_to_number(vm, &retval, &num);
+        ret = njs_value_to_number(vm, &value, &num);
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
@@ -406,7 +396,7 @@ njs_typed_array_from(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         njs_typed_array_prop_set(vm, array, i, num);
     }
 
-    njs_set_typed_array(&vm->retval, array);
+    njs_set_typed_array(retval, array);
 
     return NJS_OK;
 }
@@ -414,9 +404,9 @@ njs_typed_array_from(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 static njs_int_t
 njs_typed_array_get_this(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
-    vm->retval = args[0];
+    njs_value_assign(retval, njs_argument(args, 0));
 
     return NJS_OK;
 }
@@ -460,7 +450,7 @@ static const njs_value_t  njs_typed_array_float64_tag =
 
 static njs_int_t
 njs_typed_array_get_string_tag(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_value_t  *this;
 
@@ -479,11 +469,12 @@ njs_typed_array_get_string_tag(njs_vm_t *vm, njs_value_t *args,
     this = njs_argument(args, 0);
 
     if (!njs_is_typed_array(this)) {
-        njs_set_undefined(&vm->retval);
+        njs_set_undefined(retval);
         return NJS_OK;
     }
 
-    vm->retval = *tags[njs_typed_array_index(njs_typed_array(this)->type)];
+    njs_value_assign(retval,
+                     tags[njs_typed_array_index(njs_typed_array(this)->type)]);
 
     return NJS_OK;
 }
@@ -491,7 +482,7 @@ njs_typed_array_get_string_tag(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_length(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     uint32_t           length;
     njs_value_t        *this;
@@ -511,7 +502,7 @@ njs_typed_array_prototype_length(njs_vm_t *vm, njs_value_t *args,
         length = 0;
     }
 
-    njs_set_number(&vm->retval, length);
+    njs_set_number(retval, length);
 
     return NJS_OK;
 }
@@ -519,7 +510,7 @@ njs_typed_array_prototype_length(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_buffer(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_value_t        *this;
     njs_typed_array_t  *array;
@@ -533,7 +524,7 @@ njs_typed_array_prototype_buffer(njs_vm_t *vm, njs_value_t *args,
 
     array = njs_typed_array(this);
 
-    njs_set_array_buffer(&vm->retval, njs_typed_array_buffer(array));
+    njs_set_array_buffer(retval, njs_typed_array_buffer(array));
 
     return NJS_OK;
 }
@@ -541,7 +532,7 @@ njs_typed_array_prototype_buffer(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_byte_length(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     size_t             byte_length;
     njs_value_t        *this;
@@ -566,7 +557,7 @@ njs_typed_array_prototype_byte_length(njs_vm_t *vm, njs_value_t *args,
         byte_length = 0;
     }
 
-    njs_set_number(&vm->retval, byte_length);
+    njs_set_number(retval, byte_length);
 
     return NJS_OK;
 }
@@ -574,7 +565,7 @@ njs_typed_array_prototype_byte_length(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_byte_offset(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     size_t             byte_offset;
     njs_value_t        *this;
@@ -599,7 +590,7 @@ njs_typed_array_prototype_byte_offset(njs_vm_t *vm, njs_value_t *args,
         byte_offset = 0;
     }
 
-    njs_set_number(&vm->retval, byte_offset);
+    njs_set_number(retval, byte_offset);
 
     return NJS_OK;
 }
@@ -688,7 +679,7 @@ njs_typed_array_set_value(njs_vm_t *vm, njs_typed_array_t *array,
 
 static njs_int_t
 njs_typed_array_prototype_set(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     double              num;
     int64_t             i, length, src_length, offset;
@@ -791,7 +782,7 @@ njs_typed_array_prototype_set(njs_vm_t *vm, njs_value_t *args,
         }
     }
 
-    njs_set_undefined(&vm->retval);
+    njs_set_undefined(retval);
 
     return NJS_OK;
 }
@@ -799,7 +790,7 @@ njs_typed_array_prototype_set(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_fill(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     float               f32;
     int8_t              i8;
@@ -858,7 +849,7 @@ njs_typed_array_prototype_fill(njs_vm_t *vm, njs_value_t *args,
         return NJS_ERROR;
     }
 
-    njs_set_typed_array(&vm->retval, array);
+    njs_set_typed_array(retval, array);
 
     offset = array->offset;
 
@@ -934,7 +925,7 @@ njs_typed_array_prototype_fill(njs_vm_t *vm, njs_value_t *args,
 
 njs_int_t
 njs_typed_array_prototype_slice(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t copy)
+    njs_uint_t nargs, njs_index_t copy, njs_value_t *retval)
 {
     int64_t             start, end, count, offset;
     uint32_t            i, element_size, length;
@@ -988,7 +979,7 @@ njs_typed_array_prototype_slice(njs_vm_t *vm, njs_value_t *args,
 
         ret = njs_typed_array_species_create(vm, this,
                                              njs_value_arg(arguments), 1,
-                                             &vm->retval);
+                                             retval);
         if (njs_slow_path(ret != NJS_OK)) {
             return ret;
         }
@@ -1002,7 +993,7 @@ njs_typed_array_prototype_slice(njs_vm_t *vm, njs_value_t *args,
             return NJS_ERROR;
         }
 
-        new_array = njs_typed_array(&vm->retval);
+        new_array = njs_typed_array(retval);
         new_buffer = njs_typed_array_buffer(new_array);
         element_size = njs_typed_array_element_size(array->type);
 
@@ -1030,13 +1021,13 @@ njs_typed_array_prototype_slice(njs_vm_t *vm, njs_value_t *args,
     njs_set_number(&arguments[2], njs_max(end - start, 0));
 
     return njs_typed_array_species_create(vm, this, njs_value_arg(arguments), 3,
-                                          &vm->retval);
+                                          retval);
 }
 
 
 static njs_int_t
 njs_typed_array_prototype_copy_within(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     int64_t             length, to, from, final, count;
     uint32_t            element_size;
@@ -1090,7 +1081,7 @@ njs_typed_array_prototype_copy_within(njs_vm_t *vm, njs_value_t *args,
 
     final = (final < 0) ? njs_max(final + length, 0) : njs_min(final, length);
 
-    njs_set_typed_array(&vm->retval, array);
+    njs_set_typed_array(retval, array);
 
     count = njs_min(final - from, length - to);
 
@@ -1117,14 +1108,14 @@ njs_typed_array_prototype_copy_within(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t type)
+    njs_uint_t nargs, njs_index_t type, njs_value_t *retval)
 {
     double              val;
     int64_t             i, length;
     njs_int_t           ret;
     njs_arr_t           results;
     njs_value_t         *this, *this_arg, *r;
-    njs_value_t         arguments[4], retval;
+    njs_value_t         arguments[4], value;
     njs_function_t      *function;
     njs_typed_array_t   *array, *dst;
     njs_array_buffer_t  *buffer;
@@ -1155,12 +1146,12 @@ njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
     case NJS_ARRAY_MAP:
         njs_set_number(&arguments[0], length);
         ret = njs_typed_array_species_create(vm, this, njs_value_arg(arguments),
-                                             1, &retval);
+                                             1, &value);
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
 
-        dst = njs_typed_array(&retval);
+        dst = njs_typed_array(&value);
         break;
 
     case NJS_ARRAY_FILTER:
@@ -1184,15 +1175,15 @@ njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
         njs_set_number(&arguments[2], i);
         njs_set_typed_array(&arguments[3], array);
 
-        ret = njs_function_apply(vm, function, arguments, 4, &vm->retval);
+        ret = njs_function_apply(vm, function, arguments, 4, retval);
         if (njs_slow_path(ret != NJS_OK)) {
             goto exception;
         }
 
         switch (type) {
         case NJS_ARRAY_EVERY:
-            if (!njs_is_true(&vm->retval)) {
-                vm->retval = njs_value_false;
+            if (!njs_is_true(retval)) {
+                njs_set_boolean(retval, 0);
                 goto done;
             }
 
@@ -1204,28 +1195,28 @@ njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
         case NJS_ARRAY_SOME:
         case NJS_ARRAY_FIND:
         case NJS_ARRAY_FIND_INDEX:
-            if (!njs_is_true(&vm->retval)) {
+            if (!njs_is_true(retval)) {
                 continue;
             }
 
             switch (type) {
             case NJS_ARRAY_SOME:
-                vm->retval = njs_value_true;
+                njs_set_boolean(retval, 1);
                 break;
 
             case NJS_ARRAY_FIND:
-                njs_set_number(&vm->retval, val);
+                njs_set_number(retval, val);
                 break;
 
             default:
-                njs_set_number(&vm->retval, i);
+                njs_set_number(retval, i);
                 break;
             }
 
             goto done;
 
         case NJS_ARRAY_MAP:
-            ret = njs_typed_array_set_value(vm, dst, i, &vm->retval);
+            ret = njs_typed_array_set_value(vm, dst, i, retval);
             if (njs_slow_path(ret != NJS_OK)) {
                 goto exception;
             }
@@ -1235,7 +1226,7 @@ njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
         default:
             /* NJS_ARRAY_FILTER. */
 
-            if (!njs_is_true(&vm->retval)) {
+            if (!njs_is_true(retval)) {
                 continue;
             }
 
@@ -1252,20 +1243,20 @@ njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
 
     switch (type) {
     case NJS_ARRAY_EVERY:
-        vm->retval = njs_value_true;
+        njs_set_boolean(retval, 1);
         break;
 
     case NJS_ARRAY_SOME:
-        vm->retval = njs_value_false;
+        njs_set_boolean(retval, 0);
         break;
 
     case NJS_ARRAY_FOR_EACH:
     case NJS_ARRAY_FIND:
-        njs_set_undefined(&vm->retval);
+        njs_set_undefined(retval);
         break;
 
     case NJS_ARRAY_FIND_INDEX:
-        njs_set_number(&vm->retval, -1);
+        njs_set_number(retval, -1);
         break;
 
     case NJS_ARRAY_MAP:
@@ -1275,12 +1266,12 @@ njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
             njs_set_number(&arguments[0], results.items);
             ret = njs_typed_array_species_create(vm, this,
                                                  njs_value_arg(arguments),
-                                                 1, &retval);
+                                                 1, &value);
             if (njs_slow_path(ret != NJS_OK)) {
                 goto exception;
             }
 
-            dst = njs_typed_array(&retval);
+            dst = njs_typed_array(&value);
 
             i = 0;
 
@@ -1293,7 +1284,7 @@ njs_typed_array_prototype_iterator(njs_vm_t *vm, njs_value_t *args,
             }
         }
 
-        njs_set_typed_array(&vm->retval, dst);
+        njs_set_typed_array(retval, dst);
         break;
     }
 
@@ -1311,7 +1302,7 @@ exception:
 
 static njs_int_t
 njs_typed_array_prototype_index_of(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t type)
+    njs_uint_t nargs, njs_index_t type, njs_value_t *retval)
 {
     double              v;
     int64_t             i, i64, from, to, index, increment, offset, length;
@@ -1523,10 +1514,10 @@ done:
     /* Default values. */
 
     if (type & 1) {
-        njs_set_boolean(&vm->retval, index != -1);
+        njs_set_boolean(retval, index != -1);
 
     } else {
-        njs_set_number(&vm->retval, index);
+        njs_set_number(retval, index);
     }
 
     return NJS_OK;
@@ -1535,7 +1526,7 @@ done:
 
 static njs_int_t
 njs_typed_array_prototype_reduce(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t right)
+    njs_uint_t nargs, njs_index_t right, njs_value_t *retval)
 {
     int64_t             i, from, to, increment, length;
     njs_int_t           ret;
@@ -1610,7 +1601,7 @@ njs_typed_array_prototype_reduce(njs_vm_t *vm, njs_value_t *args,
         }
     }
 
-    vm->retval = accumulator;
+    njs_value_assign(retval, &accumulator);
 
     return NJS_OK;
 }
@@ -1618,15 +1609,15 @@ njs_typed_array_prototype_reduce(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_reverse(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t to_reversed, njs_value_t *retval)
 {
     double              *f64;
     uint8_t             *u8;
     int64_t             i, length;
     uint16_t            *u16;
     uint32_t            *u32;
-    njs_value_t         *this;
-    njs_typed_array_t   *array;
+    njs_value_t         *this, arguments[1];
+    njs_typed_array_t   *array, *self;
     njs_array_buffer_t  *buffer;
 
     this = njs_argument(args, 0);
@@ -1636,8 +1627,24 @@ njs_typed_array_prototype_reverse(njs_vm_t *vm, njs_value_t *args,
     }
 
     array = njs_typed_array(this);
-    length = njs_typed_array_length(array);
+    if (njs_slow_path(njs_is_detached_buffer(array->buffer))) {
+        njs_type_error(vm, "detached buffer");
+        return NJS_ERROR;
+    }
 
+    if (to_reversed) {
+        self = array;
+        njs_set_number(&arguments[0], njs_typed_array_length(self));
+        array = njs_typed_array_alloc(vm, arguments, 1, 0, self->type);
+        if (njs_slow_path(array == NULL)) {
+            return NJS_ERROR;
+        }
+
+        memcpy(&array->buffer->u.u8[0], &self->buffer->u.u8[0],
+               self->byte_length);
+    }
+
+    length = njs_typed_array_length(array);
     buffer = njs_typed_array_writable(vm, array);
     if (njs_slow_path(buffer == NULL)) {
         return NJS_ERROR;
@@ -1687,7 +1694,7 @@ njs_typed_array_prototype_reverse(njs_vm_t *vm, njs_value_t *args,
         }
     }
 
-    njs_set_typed_array(&vm->retval, array);
+    njs_set_typed_array(retval, array);
 
     return NJS_OK;
 }
@@ -1906,13 +1913,13 @@ exception:
 
 static njs_int_t
 njs_typed_array_prototype_sort(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t to_sorted, njs_value_t *retval)
 {
     u_char                      *base, *orig;
     int64_t                     length;
     uint32_t                    element_size;
-    njs_value_t                 *this, *comparefn;
-    njs_typed_array_t           *array;
+    njs_value_t                 *this, *comparefn, arguments[1];
+    njs_typed_array_t           *array, *self;
     njs_array_buffer_t          *buffer;
     njs_typed_array_cmp_t       cmp;
     njs_typed_array_sort_ctx_t  ctx;
@@ -1929,6 +1936,18 @@ njs_typed_array_prototype_sort(njs_vm_t *vm, njs_value_t *args,
         return NJS_ERROR;
     }
 
+    if (to_sorted) {
+        self = array;
+        njs_set_number(&arguments[0], njs_typed_array_length(self));
+        array = njs_typed_array_alloc(vm, arguments, 1, 0, self->type);
+        if (njs_slow_path(array == NULL)) {
+            return NJS_ERROR;
+        }
+
+        memcpy(&array->buffer->u.u8[0], &self->buffer->u.u8[0],
+               self->byte_length);
+    }
+
     ctx.vm = vm;
     ctx.buffer = array->buffer;
     ctx.exception = 0;
@@ -2028,7 +2047,7 @@ njs_typed_array_prototype_sort(njs_vm_t *vm, njs_value_t *args,
         return NJS_ERROR;
     }
 
-    njs_set_typed_array(&vm->retval, array);
+    njs_set_typed_array(retval, array);
 
     return NJS_OK;
 }
@@ -2076,7 +2095,7 @@ njs_typed_array_to_chain(njs_vm_t *vm, njs_chb_t *chain,
 
 static njs_int_t
 njs_typed_array_prototype_join(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     u_char             *p;
     size_t             size, length, arr_length;
@@ -2114,7 +2133,7 @@ njs_typed_array_prototype_join(njs_vm_t *vm, njs_value_t *args,
     }
 
     if (arr_length == 0) {
-        vm->retval = njs_string_empty;
+        njs_value_assign(retval, &njs_string_empty);
         return NJS_OK;
     }
 
@@ -2128,7 +2147,7 @@ njs_typed_array_prototype_join(njs_vm_t *vm, njs_value_t *args,
     length = njs_typed_array_to_chain(vm, &chain, array, separator);
     size = njs_chb_size(&chain);
 
-    p = njs_string_alloc(vm, &vm->retval, size, length);
+    p = njs_string_alloc(vm, retval, size, length);
     if (njs_slow_path(p == NULL)) {
         return NJS_ERROR;
     }
@@ -2142,7 +2161,7 @@ njs_typed_array_prototype_join(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_typed_array_prototype_iterator_obj(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t kind)
+    njs_uint_t nargs, njs_index_t kind, njs_value_t *retval)
 {
     njs_value_t        *this;
     njs_typed_array_t  *array;
@@ -2159,13 +2178,13 @@ njs_typed_array_prototype_iterator_obj(njs_vm_t *vm, njs_value_t *args,
         return NJS_ERROR;
     }
 
-    return njs_array_iterator_create(vm, this, &vm->retval, kind);
+    return njs_array_iterator_create(vm, this, retval, kind);
 }
 
 
 static njs_int_t
 njs_typed_array_constructor_intrinsic(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t unused)
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval)
 {
     njs_type_error(vm, "Abstract class TypedArray not directly constructable");
 
@@ -2175,10 +2194,10 @@ njs_typed_array_constructor_intrinsic(njs_vm_t *vm, njs_value_t *args,
 
 static const njs_object_prop_t  njs_typed_array_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("TypedArray"),
-
     NJS_DECLARE_PROP_LENGTH(0),
 
+    NJS_DECLARE_PROP_NAME("TypedArray"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     {
@@ -2293,6 +2312,11 @@ static const njs_object_prop_t  njs_typed_array_prototype_properties[] =
 
     NJS_DECLARE_PROP_NATIVE("subarray", njs_typed_array_prototype_slice, 2, 0),
 
+    NJS_DECLARE_PROP_NATIVE("toReversed", njs_typed_array_prototype_reverse, 0,
+                            1),
+
+    NJS_DECLARE_PROP_NATIVE("toSorted", njs_typed_array_prototype_sort, 1, 1),
+
     NJS_DECLARE_PROP_NATIVE("toString", njs_array_prototype_to_string, 0, 0),
 
     NJS_DECLARE_PROP_NATIVE("values", njs_typed_array_prototype_iterator_obj,
@@ -2325,7 +2349,7 @@ const njs_object_type_init_t  njs_typed_array_type_init = {
 
 static njs_int_t
 njs_data_view_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
-    njs_index_t unused)
+    njs_index_t unused, njs_value_t *retval)
 {
     uint64_t            size, offset;
     njs_int_t           ret;
@@ -2389,11 +2413,11 @@ njs_data_view_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
     njs_lvlhsh_init(&view->object.hash);
     njs_lvlhsh_init(&view->object.shared_hash);
-    view->object.__proto__ = &vm->prototypes[view->type].object;
+    view->object.__proto__ = njs_vm_proto(vm, view->type);
     view->object.type = NJS_DATA_VIEW;
     view->object.extensible = 1;
 
-    njs_set_data_view(&vm->retval, view);
+    njs_set_data_view(retval, view);
 
     return NJS_OK;
 
@@ -2407,10 +2431,10 @@ memory_error:
 
 static const njs_object_prop_t  njs_data_view_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("DataView"),
-
     NJS_DECLARE_PROP_LENGTH(1),
 
+    NJS_DECLARE_PROP_NAME("DataView"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 };
 
@@ -2423,7 +2447,7 @@ static const njs_object_init_t  njs_data_view_constructor_init = {
 
 static njs_int_t
 njs_data_view_prototype_get(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t type)
+    njs_uint_t nargs, njs_index_t type, njs_value_t *retval)
 {
     double              v;
     uint32_t            u32;
@@ -2535,7 +2559,7 @@ njs_data_view_prototype_get(njs_vm_t *vm, njs_value_t *args,
         v = conv_f64.f;
     }
 
-    njs_set_number(&vm->retval, v);
+    njs_set_number(retval, v);
 
     return NJS_OK;
 }
@@ -2543,7 +2567,7 @@ njs_data_view_prototype_get(njs_vm_t *vm, njs_value_t *args,
 
 static njs_int_t
 njs_data_view_prototype_set(njs_vm_t *vm, njs_value_t *args,
-    njs_uint_t nargs, njs_index_t type)
+    njs_uint_t nargs, njs_index_t type, njs_value_t *retval)
 {
     double              v;
     uint8_t             *u8;
@@ -2644,7 +2668,7 @@ njs_data_view_prototype_set(njs_vm_t *vm, njs_value_t *args,
         njs_set_u64(u8, conv_f64.u);
     }
 
-    njs_set_undefined(&vm->retval);
+    njs_set_undefined(retval);
 
     return NJS_OK;
 }
@@ -2737,10 +2761,10 @@ const njs_object_type_init_t  njs_data_view_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_u8_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Uint8Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Uint8Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 1),
@@ -2832,10 +2856,10 @@ const njs_object_type_init_t  njs_typed_array_u8clamped_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_i8_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Int8Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Int8Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 1),
@@ -2877,10 +2901,10 @@ const njs_object_type_init_t  njs_typed_array_i8_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_u16_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Uint16Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Uint16Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 2),
@@ -2922,10 +2946,10 @@ const njs_object_type_init_t  njs_typed_array_u16_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_i16_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Int16Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Int16Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 2),
@@ -2967,10 +2991,10 @@ const njs_object_type_init_t  njs_typed_array_i16_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_u32_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Uint32Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Uint32Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 4),
@@ -3012,10 +3036,10 @@ const njs_object_type_init_t  njs_typed_array_u32_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_i32_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Int32Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Int32Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 4),
@@ -3057,10 +3081,10 @@ const njs_object_type_init_t  njs_typed_array_i32_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_f32_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Float32Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Float32Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 4),
@@ -3102,10 +3126,10 @@ const njs_object_type_init_t  njs_typed_array_f32_type_init = {
 
 static const njs_object_prop_t  njs_typed_array_f64_constructor_props[] =
 {
-    NJS_DECLARE_PROP_NAME("Float64Array"),
-
     NJS_DECLARE_PROP_LENGTH(3),
 
+    NJS_DECLARE_PROP_NAME("Float64Array"),
+
     NJS_DECLARE_PROP_HANDLER("prototype", njs_object_prototype_create, 0, 0, 0),
 
     NJS_DECLARE_PROP_LVALUE("BYTES_PER_ELEMENT", njs_value(NJS_NUMBER, 1, 8),
