@@ -466,6 +466,15 @@ njs_value_function_set(njs_value_t *value, njs_function_t *function)
 }
 
 
+void
+njs_value_external_set(njs_value_t *value, njs_external_ptr_t external)
+{
+    njs_assert(njs_value_is_external(value, NJS_PROTO_ID_ANY));
+
+    njs_object_data(value) = external;
+}
+
+
 uint8_t
 njs_value_bool(const njs_value_t *value)
 {
@@ -487,6 +496,32 @@ njs_value_function(const njs_value_t *value)
 }
 
 
+njs_function_native_t
+njs_value_native_function(const njs_value_t *value)
+{
+    njs_function_t  *function;
+
+    if (njs_is_function(value)) {
+        function = njs_function(value);
+
+        if (function->native) {
+            return function->u.native;
+        }
+    }
+
+    return NULL;
+}
+
+
+njs_external_ptr_t
+njs_value_external(const njs_value_t *value)
+{
+    njs_assert(njs_value_is_external(value, NJS_PROTO_ID_ANY));
+
+    return njs_object_data(value);
+}
+
+
 njs_int_t
 njs_value_is_null(const njs_value_t *value)
 {
@@ -552,6 +587,20 @@ njs_value_is_object(const njs_value_t *value)
 }
 
 
+njs_int_t
+njs_value_is_error(const njs_value_t *value)
+{
+    return njs_is_error(value);
+}
+
+
+njs_int_t
+njs_value_is_external(const njs_value_t *value, njs_int_t proto_id)
+{
+    return njs_is_object_data(value, njs_make_tag(proto_id));
+}
+
+
 njs_int_t
 njs_value_is_array(const njs_value_t *value)
 {
@@ -573,6 +622,13 @@ njs_value_is_buffer(const njs_value_t *value)
 }
 
 
+njs_int_t
+njs_value_is_data_view(const njs_value_t *value)
+{
+    return njs_is_data_view(value);
+}
+
+
 /*
  * ES5.1, 8.12.1: [[GetOwnProperty]], [[GetProperty]].
  * The njs_property_query() returns values
@@ -607,7 +663,7 @@ njs_property_query(njs_vm_t *vm, njs_property_query_t *pq, njs_value_t *value,
     case NJS_NUMBER:
     case NJS_SYMBOL:
         index = njs_primitive_prototype_index(value->type);
-        obj = &vm->prototypes[index].object;
+        obj = njs_vm_proto(vm, index);
         break;
 
     case NJS_STRING:
@@ -768,7 +824,7 @@ njs_object_property_query(njs_vm_t *vm, njs_property_query_t *pq,
             }
 
             if (pq->own) {
-                pq->own_whiteout = prop;
+                pq->own_whiteout = &proto->hash;
             }
 
         } else {
@@ -864,7 +920,7 @@ njs_array_property_query(njs_vm_t *vm, njs_property_query_t *pq,
             }
 
             if (pq->own) {
-                pq->own_whiteout = prop;
+                pq->own_whiteout = &array->object.hash;
             }
 
             return NJS_DECLINED;
@@ -1189,8 +1245,11 @@ njs_value_property_set(njs_vm_t *vm, njs_value_t *value, njs_value_t *key,
     uint32_t              index;
     njs_int_t             ret;
     njs_array_t           *array;
+    njs_value_t           retval;
+    njs_flathsh_elt_t     *elt;
     njs_object_prop_t     *prop;
     njs_typed_array_t     *tarray;
+    njs_flathsh_descr_t   *h;
     njs_property_query_t  pq;
 
     static const njs_str_t  length_key = njs_str("length");
@@ -1264,7 +1323,7 @@ slow_path:
         } else {
             if (njs_prop_setter(prop) != NULL) {
                 return njs_function_call(vm, njs_prop_setter(prop),
-                                         value, setval, 1, &vm->retval);
+                                         value, setval, 1, &retval);
             }
 
             njs_key_string_get(vm, &pq.key,  &pq.lhq.key);
@@ -1275,7 +1334,7 @@ slow_path:
         }
 
         if (prop->type == NJS_PROPERTY_HANDLER) {
-            ret = njs_prop_handler(prop)(vm, prop, value, setval, &vm->retval);
+            ret = njs_prop_handler(prop)(vm, prop, value, setval, &retval);
             if (njs_slow_path(ret != NJS_DECLINED)) {
                 return ret;
             }
@@ -1320,12 +1379,44 @@ slow_path:
 
     case NJS_DECLINED:
         if (njs_slow_path(pq.own_whiteout != NULL)) {
-            /* Previously deleted property. */
+
+            /*
+             * Previously deleted property.
+             *
+             * delete it, and then
+             * insert it again as new one to preserve insertion order.
+             */
+
             if (!njs_object(value)->extensible) {
                 goto fail;
             }
 
-            prop = pq.own_whiteout;
+            pq.lhq.pool = vm->mem_pool;
+
+            int rc = njs_lvlhsh_delete(pq.own_whiteout, &pq.lhq);
+            if (rc != NJS_OK) {
+                return NJS_ERROR;
+            }
+
+            h = pq.own_whiteout->slot;
+
+            if (h == NULL) {
+                h = njs_flathsh_new(&pq.lhq);
+                if (njs_slow_path(h == NULL)) {
+                    return NJS_ERROR;
+                }
+
+                pq.own_whiteout->slot = h;
+            }
+
+            elt = njs_flathsh_add_elt(pq.own_whiteout, &pq.lhq);
+            if (njs_slow_path(elt == NULL)) {
+                return NJS_ERROR;
+            }
+
+            elt->value = (&pq.lhq)->value;
+
+            prop = (njs_object_prop_t *) elt->value;
 
             prop->type = NJS_PROPERTY;
             prop->enumerable = 1;
@@ -1497,6 +1588,7 @@ slow_path:
     }
 
     prop->type = NJS_WHITEOUT;
+    prop->enum_in_object_hash = 1;
 
     return NJS_OK;
 }
@@ -1588,6 +1680,23 @@ njs_primitive_value_to_chain(njs_vm_t *vm, njs_chb_t *chain,
 }
 
 
+njs_int_t
+njs_value_to_integer(njs_vm_t *vm, njs_value_t *value, int64_t *dst)
+{
+    double     num;
+    njs_int_t  ret;
+
+    ret = njs_value_to_number(vm, value, &num);
+    if (njs_slow_path(ret != NJS_OK)) {
+        return ret;
+    }
+
+    *dst = njs_number_to_integer(num);
+
+    return NJS_OK;
+}
+
+
 njs_int_t
 njs_value_to_object(njs_vm_t *vm, njs_value_t *value)
 {
@@ -1631,6 +1740,25 @@ njs_symbol_conversion_failed(njs_vm_t *vm, njs_bool_t to_string)
 }
 
 
+njs_int_t
+njs_value_construct(njs_vm_t *vm, njs_value_t *constructor, njs_value_t *args,
+    njs_uint_t nargs, njs_value_t *retval)
+{
+    njs_value_t   this;
+    njs_object_t  *object;
+
+    object = njs_function_new_object(vm, constructor);
+    if (njs_slow_path(object == NULL)) {
+        return NJS_ERROR;
+    }
+
+    njs_set_object(&this, object);
+
+    return njs_function_call2(vm, njs_function(constructor), &this, args,
+                              nargs, retval, 1);
+}
+
+
 njs_int_t
 njs_value_species_constructor(njs_vm_t *vm, njs_value_t *object,
     njs_value_t *default_constructor, njs_value_t *dst)
