@@ -88,7 +88,6 @@ typedef struct njs_regexp_s           njs_regexp_t;
 typedef struct njs_date_s             njs_date_t;
 typedef struct njs_object_value_s     njs_promise_t;
 typedef struct njs_property_next_s    njs_property_next_t;
-typedef struct njs_object_init_s      njs_object_init_t;
 
 
 union njs_value_s {
@@ -305,12 +304,12 @@ typedef union {
 } njs_object_prototype_t;
 
 
-typedef struct {
+struct njs_object_type_init_s {
     njs_function_t            constructor;
     const njs_object_init_t   *constructor_props;
     const njs_object_init_t   *prototype_props;
     njs_object_prototype_t    prototype_value;
-} njs_object_type_init_t;
+};
 
 
 typedef enum {
@@ -380,6 +379,7 @@ struct njs_object_prop_s {
 #define njs_prop_setter(_p)     (_p)->u.accessor.setter
 
     njs_object_prop_type_t      type:8;          /* 3 bits */
+    njs_object_prop_type_t      enum_in_object_hash:8; /* 3 bits */
 
     njs_object_attribute_t      writable:8;      /* 2 bits */
     njs_object_attribute_t      enumerable:8;    /* 2 bits */
@@ -396,8 +396,8 @@ typedef struct {
     njs_object_prop_t           scratch;
 
     njs_value_t                 key;
+    njs_lvlhsh_t                *own_whiteout;
 
-    njs_object_prop_t           *own_whiteout;
     uint8_t                     temp;
     uint8_t                     own;
 } njs_property_query_t;
@@ -626,9 +626,8 @@ typedef struct {
     ((value)->type >= NJS_OBJECT)
 
 
-#define njs_has_prototype(vm, value, proto)                                   \
-    (((njs_object_prototype_t *)                                              \
-        njs_object(value)->__proto__ - (vm)->prototypes) == proto)
+#define njs_has_prototype(vm, value, proto_id)                                \
+    (njs_object(value)->__proto__ == njs_vm_proto(vm, proto_id))
 
 
 #define njs_is_object_value(value)                                            \
@@ -1091,6 +1090,8 @@ njs_int_t njs_value_to_object(njs_vm_t *vm, njs_value_t *value);
 
 void njs_symbol_conversion_failed(njs_vm_t *vm, njs_bool_t to_string);
 
+njs_int_t njs_value_construct(njs_vm_t *vm, njs_value_t *constructor,
+    njs_value_t *args, njs_uint_t nargs, njs_value_t *retval);
 njs_int_t njs_value_species_constructor(njs_vm_t *vm, njs_value_t *object,
     njs_value_t *default_constructor, njs_value_t *dst);
 
