@@ -8,8 +8,8 @@
 #include <njs_main.h>
 
 
-static njs_int_t njs_vm_init(njs_vm_t *vm);
 static njs_int_t njs_vm_handle_events(njs_vm_t *vm);
+static njs_int_t njs_vm_protos_init(njs_vm_t *vm, njs_value_t *global);
 
 
 const njs_str_t  njs_entry_empty =          njs_str("");
@@ -78,16 +78,48 @@ njs_vm_create(njs_vm_opt_t *options)
     vm->trace.size = 2048;
     vm->trace.data = vm;
 
-    njs_set_undefined(&vm->retval);
-
     if (options->init) {
-        ret = njs_vm_init(vm);
+        ret = njs_vm_runtime_init(vm);
         if (njs_slow_path(ret != NJS_OK)) {
             return NULL;
         }
     }
 
     for (i = 0; njs_modules[i] != NULL; i++) {
+        if (njs_modules[i]->preinit == NULL) {
+            continue;
+        }
+
+        ret = njs_modules[i]->preinit(vm);
+        if (njs_slow_path(ret != NJS_OK)) {
+            return NULL;
+        }
+    }
+
+    if (options->addons != NULL) {
+        addons = options->addons;
+        for (i = 0; addons[i] != NULL; i++) {
+            if (addons[i]->preinit == NULL) {
+                continue;
+            }
+
+            ret = addons[i]->preinit(vm);
+            if (njs_slow_path(ret != NJS_OK)) {
+                return NULL;
+            }
+        }
+    }
+
+    ret = njs_vm_protos_init(vm, &vm->global_value);
+    if (njs_slow_path(ret != NJS_OK)) {
+        return NULL;
+    }
+
+    for (i = 0; njs_modules[i] != NULL; i++) {
+        if (njs_modules[i]->init == NULL) {
+            continue;
+        }
+
         ret = njs_modules[i]->init(vm);
         if (njs_slow_path(ret != NJS_OK)) {
             return NULL;
@@ -97,6 +129,10 @@ njs_vm_create(njs_vm_opt_t *options)
     if (options->addons != NULL) {
         addons = options->addons;
         for (i = 0; addons[i] != NULL; i++) {
+            if (addons[i]->init == NULL) {
+                continue;
+            }
+
             ret = addons[i]->init(vm);
             if (njs_slow_path(ret != NJS_OK)) {
                 return NULL;
@@ -114,6 +150,51 @@ njs_vm_create(njs_vm_opt_t *options)
 }
 
 
+njs_int_t
+njs_vm_ctor_push(njs_vm_t *vm)
+{
+    njs_function_t          *ctor;
+    njs_vm_shared_t         *shared;
+    njs_object_prototype_t  *prototype;
+
+    shared = vm->shared;
+
+    if (shared->constructors == NULL) {
+        shared->constructors = njs_arr_create(vm->mem_pool,
+                                              NJS_OBJ_TYPE_MAX + 8,
+                                              sizeof(njs_function_t));
+        if (njs_slow_path(shared->constructors == NULL)) {
+            njs_memory_error(vm);
+            return -1;
+        }
+
+        shared->prototypes = njs_arr_create(vm->mem_pool,
+                                              NJS_OBJ_TYPE_MAX + 8,
+                                              sizeof(njs_object_prototype_t));
+        if (njs_slow_path(shared->prototypes == NULL)) {
+            njs_memory_error(vm);
+            return -1;
+        }
+    }
+
+    ctor = njs_arr_add(shared->constructors);
+    if (njs_slow_path(ctor == NULL)) {
+        njs_memory_error(vm);
+        return -1;
+    }
+
+    prototype = njs_arr_add(shared->prototypes);
+    if (njs_slow_path(prototype == NULL)) {
+        njs_memory_error(vm);
+        return -1;
+    }
+
+    njs_assert(shared->constructors->items == shared->prototypes->items);
+
+    return shared->constructors->items - 1;
+}
+
+
 void
 njs_vm_destroy(njs_vm_t *vm)
 {
@@ -198,7 +279,7 @@ njs_vm_compile(njs_vm_t *vm, u_char **start, u_char *end)
 
     code = njs_generate_scope(vm, &generator, scope, &njs_entry_main);
     if (njs_slow_path(code == NULL)) {
-        if (!njs_is_error(&vm->retval)) {
+        if (!njs_is_error(&vm->exception)) {
             njs_internal_error(vm, "njs_generate_scope() failed");
         }
 
@@ -346,7 +427,12 @@ njs_vm_clone(njs_vm_t *vm, njs_external_ptr_t external)
     nvm->trace.data = nvm;
     nvm->external = external;
 
-    ret = njs_vm_init(nvm);
+    ret = njs_vm_runtime_init(nvm);
+    if (njs_slow_path(ret != NJS_OK)) {
+        goto fail;
+    }
+
+    ret = njs_vm_protos_init(nvm, &nvm->global_value);
     if (njs_slow_path(ret != NJS_OK)) {
         goto fail;
     }
@@ -375,8 +461,8 @@ fail:
 }
 
 
-static njs_int_t
-njs_vm_init(njs_vm_t *vm)
+njs_int_t
+njs_vm_runtime_init(njs_vm_t *vm)
 {
     njs_int_t    ret;
     njs_frame_t  *frame;
@@ -398,11 +484,6 @@ njs_vm_init(njs_vm_t *vm)
         return NJS_ERROR;
     }
 
-    ret = njs_builtin_objects_clone(vm, &vm->global_value);
-    if (njs_slow_path(ret != NJS_OK)) {
-        return NJS_ERROR;
-    }
-
     njs_lvlhsh_init(&vm->values_hash);
     njs_lvlhsh_init(&vm->keywords_hash);
     njs_lvlhsh_init(&vm->modules_hash);
@@ -417,11 +498,112 @@ njs_vm_init(njs_vm_t *vm)
 }
 
 
+void
+njs_vm_constructors_init(njs_vm_t *vm)
+{
+    njs_uint_t    i;
+    njs_object_t  *object_prototype, *function_prototype,
+                  *typed_array_prototype, *error_prototype, *async_prototype,
+                  *typed_array_ctor, *error_ctor;
+
+    object_prototype = njs_vm_proto(vm, NJS_OBJ_TYPE_OBJECT);
+
+    for (i = NJS_OBJ_TYPE_ARRAY; i < NJS_OBJ_TYPE_NORMAL_MAX; i++) {
+        vm->prototypes[i].object.__proto__ = object_prototype;
+    }
+
+    typed_array_prototype = njs_vm_proto(vm, NJS_OBJ_TYPE_TYPED_ARRAY);
+
+    for (i = NJS_OBJ_TYPE_TYPED_ARRAY_MIN;
+         i < NJS_OBJ_TYPE_TYPED_ARRAY_MAX;
+         i++)
+    {
+        vm->prototypes[i].object.__proto__ = typed_array_prototype;
+    }
+
+    vm->prototypes[NJS_OBJ_TYPE_ARRAY_ITERATOR].object.__proto__ =
+                                       njs_vm_proto(vm, NJS_OBJ_TYPE_ITERATOR);
+
+    vm->prototypes[NJS_OBJ_TYPE_BUFFER].object.__proto__ =
+                                    njs_vm_proto(vm, NJS_OBJ_TYPE_UINT8_ARRAY);
+
+    error_prototype = njs_vm_proto(vm, NJS_OBJ_TYPE_ERROR);
+    error_prototype->__proto__ = object_prototype;
+
+    for (i = NJS_OBJ_TYPE_EVAL_ERROR; i < vm->constructors_size; i++) {
+        vm->prototypes[i].object.__proto__ = error_prototype;
+    }
+
+    function_prototype = njs_vm_proto(vm, NJS_OBJ_TYPE_FUNCTION);
+
+    async_prototype = njs_vm_proto(vm, NJS_OBJ_TYPE_ASYNC_FUNCTION);
+    async_prototype->__proto__ = function_prototype;
+
+    for (i = NJS_OBJ_TYPE_OBJECT; i < NJS_OBJ_TYPE_NORMAL_MAX; i++) {
+        vm->constructors[i].object.__proto__ = function_prototype;
+    }
+
+    typed_array_ctor = &njs_vm_ctor(vm, NJS_OBJ_TYPE_TYPED_ARRAY).object;
+
+    for (i = NJS_OBJ_TYPE_TYPED_ARRAY_MIN;
+         i < NJS_OBJ_TYPE_TYPED_ARRAY_MAX;
+         i++)
+    {
+        vm->constructors[i].object.__proto__ = typed_array_ctor;
+    }
+
+    error_ctor = &njs_vm_ctor(vm, NJS_OBJ_TYPE_ERROR).object;
+    error_ctor->__proto__ = function_prototype;
+
+    for (i = NJS_OBJ_TYPE_EVAL_ERROR; i < vm->constructors_size; i++) {
+        vm->constructors[i].object.__proto__ = error_ctor;
+    }
+}
+
+
+static njs_int_t
+njs_vm_protos_init(njs_vm_t *vm, njs_value_t *global)
+{
+    size_t  ctor_size, proto_size;
+
+    vm->constructors_size = vm->shared->constructors->items;
+
+    ctor_size = vm->constructors_size * sizeof(njs_function_t);
+    proto_size = vm->constructors_size * sizeof(njs_object_prototype_t);
+
+    vm->constructors = njs_mp_alloc(vm->mem_pool, ctor_size + proto_size);
+    if (njs_slow_path(vm->constructors == NULL)) {
+        njs_memory_error(vm);
+        return NJS_ERROR;
+    }
+
+    vm->prototypes = (njs_object_prototype_t *)
+                                     ((u_char *) vm->constructors + ctor_size);
+
+    memcpy(vm->constructors, vm->shared->constructors->start, ctor_size);
+    memcpy(vm->prototypes, vm->shared->prototypes->start, proto_size);
+
+    njs_vm_constructors_init(vm);
+
+    vm->global_object.__proto__ = njs_vm_proto(vm, NJS_OBJ_TYPE_OBJECT);
+
+    njs_set_undefined(global);
+    njs_set_object(global, &vm->global_object);
+
+    vm->string_object = vm->shared->string_object;
+    vm->string_object.__proto__ = njs_vm_proto(vm, NJS_OBJ_TYPE_STRING);
+
+    return NJS_OK;
+}
+
+
 njs_int_t
 njs_vm_call(njs_vm_t *vm, njs_function_t *function, const njs_value_t *args,
     njs_uint_t nargs)
 {
-    return njs_vm_invoke(vm, function, args, nargs, &vm->retval);
+    njs_value_t  unused;
+
+    return njs_vm_invoke(vm, function, args, nargs, &unused);
 }
 
 
@@ -442,12 +624,11 @@ njs_vm_invoke(njs_vm_t *vm, njs_function_t *function, const njs_value_t *args,
 
 
 void
-njs_vm_scopes_restore(njs_vm_t *vm, njs_native_frame_t *native,
-    njs_native_frame_t *previous)
+njs_vm_scopes_restore(njs_vm_t *vm, njs_native_frame_t *native)
 {
     njs_frame_t  *frame;
 
-    vm->top_frame = previous;
+    vm->top_frame = native->previous;
 
     if (native->function->native) {
         return;
@@ -511,6 +692,16 @@ njs_vm_posted(njs_vm_t *vm)
 }
 
 
+njs_int_t
+njs_vm_unhandled_rejection(njs_vm_t *vm)
+{
+    return vm->options.unhandled_rejection
+             == NJS_VM_OPT_UNHANDLED_REJECTION_THROW
+           && vm->promise_reason != NULL
+           && vm->promise_reason->length != 0;
+}
+
+
 njs_int_t
 njs_vm_post_event(njs_vm_t *vm, njs_vm_event_t vm_event,
     const njs_value_t *args, njs_uint_t nargs)
@@ -546,11 +737,11 @@ njs_vm_run(njs_vm_t *vm)
 
 
 njs_int_t
-njs_vm_start(njs_vm_t *vm)
+njs_vm_start(njs_vm_t *vm, njs_value_t *retval)
 {
     njs_int_t  ret;
 
-    ret = njs_vmcode_interpreter(vm, vm->start, NULL, NULL);
+    ret = njs_vmcode_interpreter(vm, vm->start, retval, NULL, NULL);
 
     return (ret == NJS_ERROR) ? NJS_ERROR : NJS_OK;
 }
@@ -656,10 +847,22 @@ njs_vm_add_path(njs_vm_t *vm, const njs_str_t *path)
 }
 
 
-njs_value_t *
-njs_vm_retval(njs_vm_t *vm)
+njs_value_t
+njs_vm_exception(njs_vm_t *vm)
 {
-    return &vm->retval;
+    njs_value_t  value;
+
+    value = vm->exception;
+    njs_set_invalid(&vm->exception);
+
+    return value;
+}
+
+
+void
+njs_vm_exception_get(njs_vm_t *vm, njs_value_t *retval)
+{
+    *retval = njs_vm_exception(vm);
 }
 
 
@@ -677,6 +880,13 @@ njs_vm_external_ptr(njs_vm_t *vm)
 }
 
 
+njs_bool_t
+njs_vm_constructor(njs_vm_t *vm)
+{
+    return vm->top_frame->ctor;
+}
+
+
 uintptr_t
 njs_vm_meta(njs_vm_t *vm, njs_uint_t index)
 {
@@ -691,10 +901,48 @@ njs_vm_meta(njs_vm_t *vm, njs_uint_t index)
 }
 
 
-void
-njs_vm_retval_set(njs_vm_t *vm, const njs_value_t *value)
+njs_vm_opt_t *
+njs_vm_options(njs_vm_t *vm)
 {
-    vm->retval = *value;
+    return &vm->options;
+}
+
+
+void
+njs_vm_throw(njs_vm_t *vm, const njs_value_t *value)
+{
+    vm->exception = *value;
+}
+
+
+void
+njs_vm_error2(njs_vm_t *vm, unsigned error_type, const char *fmt, ...)
+{
+    va_list  args;
+
+    if (error_type > (NJS_OBJ_TYPE_ERROR_MAX - NJS_OBJ_TYPE_ERROR)) {
+        return;
+    }
+
+    va_start(args, fmt);
+    error_type += NJS_OBJ_TYPE_ERROR;
+    njs_throw_error_va(vm, njs_vm_proto(vm, error_type), fmt, args);
+    va_end(args);
+}
+
+
+void
+njs_vm_error3(njs_vm_t *vm, unsigned type, const char *fmt, ...)
+{
+    va_list  args;
+
+    if (type > vm->constructors_size) {
+        return;
+    }
+
+    va_start(args, fmt);
+    njs_throw_error_va(vm, njs_vm_proto(vm, type), fmt, args);
+    va_end(args);
 }
 
 
@@ -726,7 +974,7 @@ njs_vm_value(njs_vm_t *vm, const njs_str_t *path, njs_value_t *retval)
         }
 
         ret = njs_value_property(vm, &value, &key, njs_value_arg(retval));
-        if (njs_slow_path(ret != NJS_OK)) {
+        if (njs_slow_path(ret == NJS_ERROR)) {
             return ret;
         }
 
@@ -742,21 +990,15 @@ njs_vm_value(njs_vm_t *vm, const njs_str_t *path, njs_value_t *retval)
 }
 
 
-njs_int_t
-njs_vm_bind(njs_vm_t *vm, const njs_str_t *var_name, const njs_value_t *value,
+static njs_int_t
+njs_vm_bind2(njs_vm_t *vm, const njs_str_t *var_name, njs_object_prop_t *prop,
     njs_bool_t shared)
 {
     njs_int_t           ret;
     njs_object_t        *global;
     njs_lvlhsh_t        *hash;
-    njs_object_prop_t   *prop;
     njs_lvlhsh_query_t  lhq;
 
-    prop = njs_object_prop_alloc(vm, &njs_value_undefined, value, 1);
-    if (njs_slow_path(prop == NULL)) {
-        return NJS_ERROR;
-    }
-
     ret = njs_string_new(vm, &prop->name, var_name->start, var_name->length, 0);
     if (njs_slow_path(ret != NJS_OK)) {
         return NJS_ERROR;
@@ -782,6 +1024,45 @@ njs_vm_bind(njs_vm_t *vm, const njs_str_t *var_name, const njs_value_t *value,
 }
 
 
+njs_int_t
+njs_vm_bind(njs_vm_t *vm, const njs_str_t *var_name, const njs_value_t *value,
+    njs_bool_t shared)
+{
+    njs_object_prop_t   *prop;
+
+    prop = njs_object_prop_alloc(vm, &njs_value_undefined, value, 1);
+    if (njs_slow_path(prop == NULL)) {
+        return NJS_ERROR;
+    }
+
+    return njs_vm_bind2(vm, var_name, prop, shared);
+}
+
+
+njs_int_t
+njs_vm_bind_handler(njs_vm_t *vm, const njs_str_t *var_name,
+    njs_prop_handler_t handler, uint16_t magic16, uint32_t magic32,
+    njs_bool_t shared)
+{
+    njs_object_prop_t  *prop;
+
+    prop = njs_object_prop_alloc(vm, &njs_string_empty,
+                                 &njs_value_invalid, 1);
+    if (njs_slow_path(prop == NULL)) {
+        return NJS_ERROR;
+    }
+
+    prop->type = NJS_PROPERTY_HANDLER;
+    prop->u.value.type = NJS_INVALID;
+    prop->u.value.data.truth = 1;
+    njs_prop_magic16(prop) = magic16;
+    njs_prop_magic32(prop) = magic32;
+    njs_prop_handler(prop) = handler;
+
+    return njs_vm_bind2(vm, var_name, prop, shared);
+}
+
+
 void
 njs_value_string_get(njs_value_t *value, njs_str_t *dst)
 {
@@ -881,7 +1162,6 @@ njs_int_t
 njs_vm_prop_name(njs_vm_t *vm, njs_object_prop_t *prop, njs_str_t *dst)
 {
     if (njs_slow_path(!njs_is_string(&prop->name))) {
-        njs_type_error(vm, "property name is not a string");
         return NJS_ERROR;
     }
 
@@ -891,28 +1171,10 @@ njs_vm_prop_name(njs_vm_t *vm, njs_object_prop_t *prop, njs_str_t *dst)
 }
 
 
-njs_noinline void
-njs_vm_value_error_set(njs_vm_t *vm, njs_value_t *value, const char *fmt, ...)
-{
-    va_list  args;
-    u_char   buf[NJS_MAX_ERROR_STR], *p;
-
-    p = buf;
-
-    if (fmt != NULL) {
-        va_start(args, fmt);
-        p = njs_vsprintf(buf, buf + sizeof(buf), fmt, args);
-        va_end(args);
-    }
-
-    njs_error_new(vm, value, NJS_OBJ_TYPE_ERROR, buf, p - buf);
-}
-
-
 njs_noinline void
 njs_vm_memory_error(njs_vm_t *vm)
 {
-    njs_memory_error_set(vm, &vm->retval);
+    njs_memory_error_set(vm, &vm->exception);
 }
 
 
@@ -945,6 +1207,17 @@ njs_vm_value_string(njs_vm_t *vm, njs_str_t *dst, njs_value_t *src)
 {
     njs_int_t    ret;
     njs_uint_t   exception;
+    njs_value_t  value;
+
+    if (njs_slow_path(vm->top_frame == NULL)) {
+        /* An exception was thrown during compilation. */
+        njs_vm_runtime_init(vm);
+    }
+
+    if (njs_is_valid(&vm->exception)) {
+        value = njs_vm_exception(vm);
+        src = &value;
+    }
 
     if (njs_slow_path(src->type == NJS_NUMBER
                       && njs_number(src) == 0
@@ -968,7 +1241,7 @@ again:
 
         /* value evaluation threw an exception. */
 
-        src = &vm->retval;
+        *src = njs_vm_exception(vm);
         goto again;
     }
 
@@ -980,28 +1253,13 @@ again:
 
 
 njs_int_t
-njs_vm_retval_string(njs_vm_t *vm, njs_str_t *dst)
+njs_vm_exception_string(njs_vm_t *vm, njs_str_t *dst)
 {
-    if (vm->top_frame == NULL) {
-        /* An exception was thrown during compilation. */
+    njs_value_t  exception;
 
-        njs_vm_init(vm);
-    }
+    exception = njs_vm_exception(vm);
 
-    return njs_vm_value_string(vm, dst, &vm->retval);
-}
-
-
-njs_int_t
-njs_vm_retval_dump(njs_vm_t *vm, njs_str_t *dst, njs_uint_t indent)
-{
-    if (vm->top_frame == NULL) {
-        /* An exception was thrown during compilation. */
-
-        njs_vm_init(vm);
-    }
-
-    return njs_vm_value_dump(vm, dst, &vm->retval, 0, 1);
+    return njs_vm_value_string(vm, dst, &exception);
 }
 
 
@@ -1146,6 +1404,32 @@ njs_vm_object_prop(njs_vm_t *vm, njs_value_t *value, const njs_str_t *prop,
 }
 
 
+njs_int_t
+njs_vm_object_prop_set(njs_vm_t *vm, njs_value_t *value, const njs_str_t *prop,
+    njs_opaque_value_t *setval)
+{
+    njs_int_t    ret;
+    njs_value_t  key;
+
+    if (njs_slow_path(!njs_is_object(value))) {
+        njs_type_error(vm, "njs_vm_object_prop_set() argument is not object");
+        return NJS_ERROR;
+    }
+
+    ret = njs_vm_value_string_set(vm, &key, prop->start, prop->length);
+    if (njs_slow_path(ret != NJS_OK)) {
+        return NJS_ERROR;
+    }
+
+    ret = njs_value_property_set(vm, value, &key, njs_value_arg(setval));
+    if (njs_slow_path(ret != NJS_OK)) {
+        return NJS_ERROR;
+    }
+
+    return NJS_OK;
+}
+
+
 njs_value_t *
 njs_vm_array_prop(njs_vm_t *vm, njs_value_t *value, int64_t index,
     njs_opaque_value_t *retval)
@@ -1177,6 +1461,14 @@ njs_vm_array_prop(njs_vm_t *vm, njs_value_t *value, int64_t index,
 }
 
 
+njs_int_t
+njs_vm_object_iterate(njs_vm_t *vm, njs_iterator_args_t *args,
+    njs_iterator_handler_t handler, njs_value_t *retval)
+{
+    return njs_object_iterate(vm, args, handler, retval);
+}
+
+
 njs_value_t *
 njs_vm_array_start(njs_vm_t *vm, njs_value_t *value)
 {
@@ -1200,6 +1492,22 @@ njs_vm_array_length(njs_vm_t *vm, njs_value_t *value, int64_t *length)
 }
 
 
+njs_int_t
+njs_vm_date_alloc(njs_vm_t *vm, njs_value_t *retval, double time)
+{
+    njs_date_t  *date;
+
+    date = njs_date_alloc(vm, time);
+    if (njs_slow_path(date == NULL)) {
+        return NJS_ERROR;
+    }
+
+    njs_set_date(retval, date);
+
+    return NJS_OK;
+}
+
+
 njs_int_t
 njs_vm_value_to_string(njs_vm_t *vm, njs_str_t *dst, njs_value_t *src)
 {
@@ -1297,6 +1605,31 @@ njs_vm_value_to_c_string(njs_vm_t *vm, njs_value_t *value)
 }
 
 
+njs_int_t
+njs_value_to_string(njs_vm_t *vm, njs_value_t *dst, njs_value_t *value)
+{
+    njs_int_t    ret;
+    njs_value_t  primitive;
+
+    if (njs_slow_path(!njs_is_primitive(value))) {
+        if (njs_slow_path(njs_is_object_symbol(value))) {
+            /* should fail */
+            value = njs_object_value(value);
+
+        } else {
+            ret = njs_value_to_primitive(vm, &primitive, value, 1);
+            if (njs_slow_path(ret != NJS_OK)) {
+                return ret;
+            }
+
+            value = &primitive;
+        }
+    }
+
+    return njs_primitive_value_to_string(vm, dst, value);
+}
+
+
 njs_int_t
 njs_vm_value_to_bytes(njs_vm_t *vm, njs_str_t *dst, njs_value_t *src)
 {
@@ -1421,7 +1754,7 @@ njs_vm_value_string_copy(njs_vm_t *vm, njs_str_t *retval,
 void *
 njs_lvlhsh_alloc(void *data, size_t size)
 {
-    return njs_mp_align(data, size, size);
+    return njs_mp_align(data, NJS_MAX_ALIGNMENT, size);
 }
 
 
