@@ -76,6 +76,7 @@ typedef enum {
     NJS_OBJ_TYPE_URI_ERROR,
     NJS_OBJ_TYPE_MEMORY_ERROR,
     NJS_OBJ_TYPE_AGGREGATE_ERROR,
+#define NJS_OBJ_TYPE_ERROR_MAX         (NJS_OBJ_TYPE_AGGREGATE_ERROR)
 
     NJS_OBJ_TYPE_MAX,
 } njs_object_type_t;
@@ -95,9 +96,6 @@ enum njs_object_e {
     NJS_OBJECT_PROCESS,
     NJS_OBJECT_MATH,
     NJS_OBJECT_JSON,
-#ifdef NJS_TEST262
-    NJS_OBJECT_262,
-#endif
     NJS_OBJECT_MAX
 };
 
@@ -118,8 +116,7 @@ typedef enum {
 
 
 struct njs_vm_s {
-    /* njs_vm_t must be aligned to njs_value_t due to scratch value. */
-    njs_value_t              retval;
+    njs_value_t              exception;
 
     njs_arr_t                *paths;
     njs_arr_t                *protos;
@@ -145,13 +142,12 @@ struct njs_vm_s {
 
     njs_vm_opt_t             options;
 
-    /*
-     * The prototypes and constructors arrays must be together because
-     * they are copied from njs_vm_shared_t by single memcpy()
-     * in njs_builtin_objects_clone().
-     */
-    njs_object_prototype_t   prototypes[NJS_OBJ_TYPE_MAX];
-    njs_function_t           constructors[NJS_OBJ_TYPE_MAX];
+#define njs_vm_proto(vm, index) (&(vm)->prototypes[index].object)
+#define njs_vm_ctor(vm, index) ((vm)->constructors[index])
+
+    njs_object_prototype_t   *prototypes;
+    njs_function_t           *constructors;
+    size_t                   constructors_size;
 
     njs_function_t           *hooks[NJS_HOOK_MAX];
 
@@ -226,32 +222,34 @@ struct njs_vm_shared_s {
     njs_object_t             string_object;
     njs_object_t             objects[NJS_OBJECT_MAX];
 
-    njs_exotic_slots_t       global_slots;
+#define njs_shared_ctor(shared, index)                                       \
+    ((njs_function_t *) njs_arr_item((shared)->constructors, index))
 
-    /*
-     * The prototypes and constructors arrays must be togther because they are
-     * copied to njs_vm_t by single memcpy() in njs_builtin_objects_clone().
-     */
-    njs_object_prototype_t   prototypes[NJS_OBJ_TYPE_MAX];
-    njs_function_t           constructors[NJS_OBJ_TYPE_MAX];
+#define njs_shared_prototype(shared, index)                                  \
+    ((njs_object_prototype_t *) njs_arr_item((shared)->prototypes, index))
+
+    njs_arr_t                *constructors; /* of njs_function_t */
+    njs_arr_t                *prototypes; /* of njs_object_prototype_t */
+
+    njs_exotic_slots_t       global_slots;
 
     njs_regexp_pattern_t     *empty_regexp_pattern;
 };
 
 
-void njs_vm_scopes_restore(njs_vm_t *vm, njs_native_frame_t *frame,
-    njs_native_frame_t *previous);
+njs_int_t njs_vm_runtime_init(njs_vm_t *vm);
+njs_int_t njs_vm_ctor_push(njs_vm_t *vm);
+void njs_vm_constructors_init(njs_vm_t *vm);
+njs_value_t njs_vm_exception(njs_vm_t *vm);
+void njs_vm_scopes_restore(njs_vm_t *vm, njs_native_frame_t *frame);
 
 njs_int_t njs_builtin_objects_create(njs_vm_t *vm);
-njs_int_t njs_builtin_objects_clone(njs_vm_t *vm, njs_value_t *global);
 njs_int_t njs_builtin_match_native_function(njs_vm_t *vm,
     njs_function_t *function, njs_str_t *name);
 
 void njs_disassemble(u_char *start, u_char *end, njs_int_t count,
     njs_arr_t *lines);
 
-njs_arr_t *njs_vm_completions(njs_vm_t *vm, njs_str_t *expression);
-
 void *njs_lvlhsh_alloc(void *data, size_t size);
 void njs_lvlhsh_free(void *data, void *p, size_t size);
 
